# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



from __future__ import print_function, unicode_literals, absolute_import, division
import six
from webob import Response
from webob.exc import HTTPMethodNotAllowed

from reahl.stubble import stubclass, exempt
from reahl.tofu import scenario, Fixture
from reahl.tofu.pytestsupport import with_fixtures

from reahl.component.modelinterface import Field
from reahl.webdev.tools import Browser
from reahl.web.fw import Resource, SubResource, Widget

from reahl.web_dev.fixtures import WebFixture


@with_fixtures(WebFixture)
def test_resources(web_fixture):
    """A Resource has the responsibility to handle an HTTP Request. A Resource indicates that it can
       handle a particular HTTP method by having a method named for it. Such method should return
       a Response."""

    fixture = web_fixture

    @stubclass(Resource)
    class ResourceStub(Resource):
        called = None
        @exempt
        def handle_something(self, request):
            self.called = True
            return 'something'
        @exempt
        def handle_anotherthing(self, request):
            pass

    resource = ResourceStub()

    # Case where the HTTP method is not supported
    fixture.request.method = 'koos'
    response = resource.handle_request(fixture.request)
    assert isinstance(response, HTTPMethodNotAllowed)
    assert not resource.called
    assert response.headers['allow'] == 'anotherthing, something'

    # Case where the HTTP method is supported
    fixture.request.method = 'SOMEthing'
    response = resource.handle_request(fixture.request)
    assert resource.called
    assert response == 'something'


@with_fixtures(WebFixture)
def test_simple_sub_resources(web_fixture):
    """During their construction, Widgets can add SubResources to their View.  The SubResource
       will then be available via a special URL underneath the URL of the Widget's View."""

    fixture = web_fixture

    @stubclass(SubResource)
    class ASimpleSubResource(SubResource):
        sub_regex = 'simple_resource'
        sub_path_template = 'simple_resource'
        @exempt
        def handle_get(self, request):
            return Response()

    @stubclass(Widget)
    class WidgetWithSubResource(Widget):
        def __init__(self, view):
            super(WidgetWithSubResource, self).__init__(view)
            view.add_resource(ASimpleSubResource('uniquename'))

    wsgi_app = fixture.new_wsgi_app(child_factory=WidgetWithSubResource.factory())
    browser = Browser(wsgi_app)
    browser.open('/_uniquename_simple_resource')


@with_fixtures(WebFixture)
def test_dynamic_sub_resources(web_fixture):
    """Sometimes it is undesirable to instantiate a SubResource just for the purpose of adding it
       to a View. In such cases, A Factory can be added that will construct the SubResource
       on demand, based on a regex which may contain parameters."""
    fixture = web_fixture

    @stubclass(SubResource)
    class ParameterisedSubResource(SubResource):
        sub_regex = 'dynamic_(?P<param>[^/]+)'
        sub_path_template = 'dynamic_%(param)s'

        def __init__(self, unique_name, param):
            super(ParameterisedSubResource, self).__init__(unique_name)
            self.param = param

        @exempt
        def handle_get(self, request):
            return Response(unicode_body=six.text_type(self.param))

        @exempt
        @classmethod
        def create_resource(cls, unique_name, param=None):
            return cls(unique_name, param)

    @stubclass(Widget)
    class WidgetWithSubResource(Widget):
        def __init__(self, view):
            super(WidgetWithSubResource, self).__init__(view)
            view.add_resource_factory(ParameterisedSubResource.factory('uniquename', {'param': Field(required=True)}))

    wsgi_app = fixture.new_wsgi_app(child_factory=WidgetWithSubResource.factory())
    browser = Browser(wsgi_app)
    browser.open('/_uniquename_dynamic_one')
    assert browser.raw_html == 'one'
    browser.open('/_uniquename_dynamic_two')
    assert browser.raw_html == 'two'


@with_fixtures(WebFixture)
def test_dynamic_sub_resources_factory_args(web_fixture):
    """Such dynamic SubResources can also be created with arguments specified to its Factory (instead of
       only from the path)."""
    @stubclass(SubResource)
    class ParameterisedSubResource(SubResource):
        sub_regex = 'dynamic_(?P<path_param>[^/]+)'
        sub_path_template = 'dynamic_%(path_param)s'

        def __init__(self, unique_name, factory_arg, factory_kwarg, path_param):
            super(ParameterisedSubResource, self).__init__(unique_name)
            self.path_param = path_param
            self.factory_arg = factory_arg
            self.factory_kwarg = factory_kwarg

        @exempt
        def handle_get(self, request):
            args = '%s|%s|%s' % (self.factory_arg, self.factory_kwarg, self.path_param)
            return Response(unicode_body=six.text_type(args))

        @exempt
        @classmethod
        def create_resource(cls, unique_name, factory_arg, factory_kwarg=None, path_param=None):
            return cls(unique_name, factory_arg, factory_kwarg, path_param)

    @stubclass(Widget)
    class WidgetWithSubResource(Widget):
        def __init__(self, view):
            super(WidgetWithSubResource, self).__init__(view)
            factory = ParameterisedSubResource.factory('uniquename', {'path_param': Field(required=True)}, 'arg to factory', factory_kwarg='kwarg to factory')
            view.add_resource_factory(factory)

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(child_factory=WidgetWithSubResource.factory())
    browser = Browser(wsgi_app)
    browser.open('/__uniquename_dynamic_one')
    assert browser.raw_html == 'arg to factory|kwarg to factory|one'



@with_fixtures(WebFixture)
def test_disambiguating_between_factories(web_fixture):
    """Sometimes, Widgets may need to add more than one SubResource (or Factories) of the same type.
       Since these are of the same type, they will both match the same urls based on their sub_regex.
       For this reason, one can use the unique_name passed to a SubResource or its Factory
       to disambiguate."""
    @stubclass(SubResource)
    class ParameterisedSubResource(SubResource):
        sub_regex = 'path'
        sub_path_template = 'path'

        @exempt
        def handle_get(self, request):
            # We send back the unique_name of the factory so we can test which factory handled the request
            return Response(unicode_body=six.text_type(self.unique_name))

        @exempt
        @classmethod
        def create_resource(cls, unique_name):
            return cls(unique_name)

    @stubclass(Widget)
    class WidgetWithAmbiguousSubResources(Widget):
        def __init__(self, view):
            super(WidgetWithAmbiguousSubResources, self).__init__(view)
            factory1 = ParameterisedSubResource.factory('factory1', {})
            view.add_resource_factory(factory1)
            factory2 = ParameterisedSubResource.factory('factory2', {})
            view.add_resource_factory(factory2)

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(child_factory=WidgetWithAmbiguousSubResources.factory())
    browser = Browser(wsgi_app)
    browser.open('/__factory1_path')
    assert browser.raw_html == 'factory1'
    browser.open('/__factory2_path')
    assert browser.raw_html == 'factory2'


class UrlScenarios(Fixture):
    @scenario
    def ending_in_slash(self):
        self.view_path = '/a/'
        self.expected_path = '/a/__uniquename_sub_path'
    @scenario
    def not_ending_in_slash(self):
        self.view_path = '/a'
        self.expected_path = '/a/_uniquename_sub_path'
    @scenario
    def only_slash(self):
        self.view_path = '/'
        self.expected_path = '/__uniquename_sub_path'


@with_fixtures(WebFixture, UrlScenarios)
def test_computation_of_url(web_fixture, url_scenarios):
    """The URL of a SubResource can be different, depending on the scenario.  This is done so that
       A SubResource can unambiguously determine the URL of its parent View."""
    @stubclass(SubResource)
    class ASimpleSubResource(SubResource):
        sub_regex = 'sub_path'
        sub_path_template = 'sub_path'


    sub_resource = ASimpleSubResource('uniquename')

    # Calculating the sub_resource's URL from the context of the View
    web_fixture.request.environ['PATH_INFO'] = url_scenarios.view_path
    actual = sub_resource.get_url()
    assert actual.path == url_scenarios.expected_path

    # Calculating the URL of the parent View from the context of the SubResource
    web_fixture.request.environ['PATH_INFO'] = url_scenarios.expected_path
    actual = SubResource.get_parent_url()
    assert actual.path == url_scenarios.view_path

    # Calculating the sub_resource's URL from the context of the SubResource
    web_fixture.request.environ['PATH_INFO'] = url_scenarios.expected_path
    actual = sub_resource.get_url()
    assert actual.path == url_scenarios.expected_path



# Copyright 2015-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

""".. versionadded:: 3.2

TabbedPanels are used to save space by stacking different panels on
top of one another. While each panel has its own contents only the
top panel is visible, thus taking up visual space of only one such
panel.

"""
from __future__ import print_function, unicode_literals, absolute_import, division

# noinspection PyUnresolvedReferences
import six

from reahl.component.modelinterface import exposed, Field
from reahl.web.fw import Widget
from reahl.web.ui import Div, ActiveStateAttributes, DelegatedAttributes
from reahl.web.bootstrap.navs import Nav, TabLayout, DropdownMenu


class TabbedPanel(Widget):
    """A Widget that appears as having multiple panels of content stacked 
    on top of one another -- with only one panel visible at a time.

    Each content panel appears to have a label sticking out (its
    "tab").  When the user clicks on a different Tab, its panel is
    raised to the top and thus made visible. This is done without
    refreshing the entire page, provided that JavaScript is available
    on the user agent.

    :param view: (See :class:`reahl.web.fw.Widget`)
    :keyword nav_layout: Optionally, an :class:`~reahl.web.bootstrap.navs.PillLayout` or 
                  :class:`~reahl.web.bootstrap.navs.TabLayout` for fine-grained control over 
                  how the :class:`~reahl.web.bootstrap.navs.Nav` which represents the tabs is
                  displayed.
    """
    def __init__(self, view, nav_layout=None):
        super(TabbedPanel, self).__init__(view)
        self.tabs = []
        self.nav = self.add_child(Nav(view).use_layout(nav_layout or TabLayout()))
        self.content_panel = self.add_child(Div(view))
        self.content_panel.append_class('tab-content')

    @exposed
    def query_fields(self, fields):
        fields.tab = Field(required=False, default=None)

    @property
    def active_tab_set(self):
        return self.tab is not None

    def set_active(self, tab):
        # noinspection PyAttributeOutsideInit
        self.tab = tab.default_active_tab_key

    def is_currently_open(self, tab):
        return tab.tab_key == self.tab

    def add_tab(self, tab):
        """Add a panel for the given :class:`Tab` to this TabbedPanel.

        :param tab: The :class:`Tab` to be added.
        """
        if not self.active_tab_set:
            self.set_active(tab)
        tab.set_panel(self)

        self.tabs.append(tab)
        tab.add_to_menu(self.nav)
        tab.add_contents_to(self.content_panel)

        return tab


class TabContentAttributes(DelegatedAttributes):
    def __init__(self, tab):
        super(TabContentAttributes, self).__init__()
        self.tab = tab

    def set_attributes(self, attributes):
        super(TabContentAttributes, self).set_attributes(attributes)

        attributes.set_to('data-toggle', self.tab.data_toggle)
        attributes.set_to('data-target', '#%s' % self.tab.css_id)
        attributes.set_to('aria-controls', '%s' % self.tab.css_id)
        attributes.set_to('role', 'tab')


class Tab(object):
    """
    One Tab in a :class:`TabbedPanel`, including the contents that should be displayed for it.

    :param view: (See :class:`reahl.web.fw.Widget`)
    :param title: The label Text that is displayed inside the Tab itself.
    :param tab_key: A name for this tag identifying it uniquely amongst other Tabs in the same :class:`TabbedPanel`.
    :param contents_factory: A :class:`reahl.web.fw.WidgetFactory` specifying how to create the contents of this Tab, 
                            once selected.
    """
    def __init__(self, view, title, tab_key, contents_factory):
        self.title = title
        self.tab_key = tab_key
        self.contents_factory = contents_factory
        self.view = view
        self.panel = None
        self.menu_item = None

    def get_bookmark(self, view):
        return view.as_bookmark(description=self.title, query_arguments=self.query_arguments)

    @property
    def top_level_nav(self):
        return self.panel.nav

    @property
    def data_toggle(self):
        return self.top_level_nav.layout.key

    @property
    def default_active_tab_key(self):
        return self.tab_key

    @property
    def query_arguments(self):
        return {'tab': self.tab_key}

    @property
    def contents(self):
        return self.contents_factory.create(self.view)

    def set_panel(self, tabbed_panel):
        self.panel = tabbed_panel

    @property
    def is_open(self):
        return self.panel.is_currently_open(self)

    @property
    def is_active(self):
        return self.is_open

    def add_to_menu(self, menu):
        menu_item = menu.add_bookmark(self.get_bookmark(menu.view))
        menu_item.determine_is_active_using(lambda: self.is_active)
        menu_item.a.set_css_id('nav_%s_tab' % self.css_id)
        menu_item.a.add_attribute_source(TabContentAttributes(self))
        menu_item.a.add_attribute_source(ActiveStateAttributes(menu_item, attribute_name='aria-selected', active_value='true', inactive_value='false'))
        self.menu_item = menu_item
        return menu_item

    @property
    def css_id(self):
        return 'tab_%s' % self.tab_key

    def add_contents_to(self, content_panel):
        return self.add_contents_of_tab_to(content_panel, self)

    def add_contents_of_tab_to(self, content_panel, tab):
        assert tab.menu_item, 'add the tab to a menu first before adding its contents'
        div = content_panel.add_child(Div(self.view))
        div.add_child(tab.contents)
        div.append_class('tab-pane')
        div.set_attribute('role', 'tabpanel')
        div.set_id(tab.css_id)
        div.add_attribute_source(ActiveStateAttributes(tab, active_value='active'))
        div.set_attribute('aria-labelledby', '%s' % tab.menu_item.a.css_id)
        return div


class MultiTab(Tab):
    """
    A composite tab. Instead of a single choice for the user, clicking on a MultiTab
    results in a dropdown menu with more choices for the user. These second-level
    choices are navigable tabs.

    :param view: (See :class:`reahl.web.fw.Widget`)
    :param title: (See :class:`Tab`)
    :param tab_key: (See :class:`Tab`)
    """
    def __init__(self, view, title, tab_key):
        self.tabs = []
        self.menu = DropdownMenu(view)
        super(MultiTab, self).__init__(view, title, tab_key, None)
        
    def add_tab(self, tab):
        tab.add_to_menu(self.menu)
        tab.set_panel(self.panel)
        self.tabs.append(tab)
        return tab

    @property
    def query_arguments(self):
        return {'tab': self.panel.tab}

    @property
    def first_tab(self):
        return self.tabs[0]

    @property
    def current_sub_tab(self):
        open_tab = [tab for tab in self.tabs
                    if tab.is_open]
        if len(open_tab) == 1:
            return open_tab[0]
        return self.first_tab

    def add_contents_to(self, content_panel):
        for tab in self.tabs:
            self.add_contents_of_tab_to(content_panel, tab)

    def set_panel(self, tabbed_panel):
        super(MultiTab, self).set_panel(tabbed_panel)
        for tab in self.tabs:
            tab.set_panel(tabbed_panel)

    @property
    def default_active_tab_key(self):
        return self.first_tab.tab_key

    @property
    def is_active(self):
        return self.is_open or self.current_sub_tab.is_open

    def add_to_menu(self, menu):
        menu_item = menu.add_dropdown(self.title, self.menu, query_arguments=self.query_arguments)
        menu_item.determine_is_active_using(lambda: self.is_active)
        return menu_item



# Copyright 2015-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
.. versionadded:: 3.2

Generate images on the fly.


"""
from __future__ import print_function, unicode_literals, absolute_import, division

from collections import OrderedDict

from reahl.web.fw import Url
from reahl.web.ui import Img, HTMLAttributeValueOption


class Theme(OrderedDict):
    def __init__(self):
        super(Theme, self).__init__()


class PredefinedTheme(Theme):
    """There are a few named, :class:`PredefinedTheme`\s you can choose from to
    control the look of a :class:`PlaceholderImage` .

    :param theme_name: One of: 'sky', 'vine', 'lava', 'gray', 'industrial', or 'social'.
    """
    def __init__(self, theme_name):
        super(PredefinedTheme, self).__init__()
        theme_name_option = HTMLAttributeValueOption(theme_name, theme_name is not None,
                                                     constrain_value_to=['sky', 'vine', 'lava', 'gray', 'industrial', 'social'])
        self['theme'] = theme_name_option.as_html_snippet()


class CustomTheme(Theme):
    """A :class:`CustomTheme` allows one to control all the details of
    what a :class:`PlaceholderImage` should look like.

    :keyword bg: The background-colour (a string in CSS notation)
    :keyword fg: The foreground-colour (a string in CSS notation)
    :keyword text_size: The size of generated text (an int, denoted in pts)
    :keyword text_font: The name of the font to use for generated text.
    :keyword text_align: How to align the generated text (one of 'left' or 'right')
    :keyword line_wrap: A ratio (line length to image width) at which generated text should wrap.
    :keyword outline: Draws a border and diagonals in the image.
    """
    def __init__(self, bg=None, fg=None, text_size=None, text_font=None, text_align=None, line_wrap=None, outline=None):

        super(CustomTheme, self).__init__()

        if bg is not None:
            self['bg'] = bg
        if fg is not None:
            self['fg'] = fg
        if text_size is not None:
            self['size'] = str(text_size)
        if text_font is not None:
            self['font'] = text_font
        if text_align is not None:
            text_align_option = HTMLAttributeValueOption(text_align,  text_align is not None,
                                                         constrain_value_to=['left', 'right'])
            self['align'] = text_align_option.as_html_snippet()
        if outline is not None:
            self['outline'] = outline
        if line_wrap is not None:
            self['lineWrap'] = str(line_wrap)


class PlaceholderImage(Img):
    """A :class:`PlaceholderImage` is an image generated on the client browser using SVG.

    :param view: (See :class:`~reahl.web.fw.Widget`)
    :param x: The width of the image (defaults to pixels, can be a string using CSS notation).
    :param y: The height of the image (defaults to pixels, can be a string using CSS notation).
    :keyword alt: Text to be displayed when the browser cannot handle images.
    :keyword text: Text to be generated on the image itself.
    :keyword theme: A :class:`PredefinedTheme` or :class:`CustomTheme` to control what the image should look like.
    """
    def __init__(self, view, x, y, alt=None, text=None, theme=None):

        super(PlaceholderImage, self).__init__(view, alt=alt)
        image_url = Url('holder.js/%sx%s' % (x, y))
        options = theme or Theme()
        image_url.set_query_from(options)
        if text is not None:  
            # Our query string will url-escape text. Text often has spaces 
            # that would now become +. Holder does not do the reverse, so 
            # we have to add the text as an un-escaped query string argument 
            image_url = '%s%stext=%s' % (str(image_url), '?' if not options else '&', text)
        self.set_attribute('data-src', str(image_url))









# Copyright 2015-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

""".. versionadded:: 3.2

This module contains tools for controlling how elements are positioned.

The main tool here is an invisible layout grid which changes depending
on the size of the device being used to browse. The module also contains
several other related tools.

When creating a visual layout it is often useful to arrange elements
on an invisible grid. The tools here allows building such a grid, but
with a twist: the grid can resize depending on the size of the device
that is used to view it.

User devices come in a wide range of sizes. In order to be able
to change a layout depending on the size of a device, devices are classified
into several device classes: 'xs' (extra small), 'sm' (small), 'md'(medium),
'lg' (large), 'xl' (extra large).

Whenever you need to specify a size for Bootstrap grid element, you
can specify the size that element should have *for each class of device*.

Bootstrap's grid system works on units of 1/12th the size of a given
parent width. A size for a particular device class is thus an integer
denoting a size in 1/12ths of its container's width.
"""
from __future__ import print_function, unicode_literals, absolute_import, division
import six

from collections import OrderedDict
import copy

from reahl.component.exceptions import ProgrammerError
from reahl.component.exceptions import arg_checks, IsInstance
from reahl.web.fw import Layout
from reahl.web.ui import Div, Slot


class Container(Layout):
    """A Container Layout manages the positioning of the main areas of
    a page. By default it ensures that the HTMLElement it is used with to be
    has a size that stays at a fixed width per device class and that it stays
    centered in the horizontal.

    Using a Container is compulsory if you want to make use of a
    ColumnLayout.

    :keyword fluid: If True, the container fills the entire available width.

    """
    def __init__(self, fluid=False):
        super(Container, self).__init__()
        self.fluid = fluid

    def customise_widget(self):
        container_class = 'container'
        if self.fluid:
            container_class = 'container-fluid'
        self.widget.append_class(container_class)


class DeviceClass(object):
    device_classes = ['xs', 'sm', 'md', 'lg', 'xl']

    def __init__(self, name):
        self.name = name
        if not name in self.device_classes:
            raise ProgrammerError('Invalid device class name: %s, should be one of %s' % (name, ','.join(self.device_classes)))

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    @property
    def one_smaller(self):
        index_of_one_smaller_class = self.device_classes.index(self.name) - 1
        if index_of_one_smaller_class < 0:
            return None
        return DeviceClass(self.device_classes[index_of_one_smaller_class])

    @property
    def all_smaller(self):
        return [DeviceClass(i) for i in self.device_classes[:self.device_classes.index(self.name)]]

    @classmethod
    def all_classes(cls):
        return [DeviceClass(i) for i in DeviceClass.device_classes]

    def as_combined_css_class(self, before_parts, after_parts):
        parts = before_parts + list(filter(None, [None if self.name == 'xs' else self.name])) + after_parts
        return '-'.join(parts)


class ResponsiveOption(object):
    def __init__(self, allowed_options, prefix=None, xs=None, sm=None, md=None, lg=None, xl=None):
        self.prefix = prefix
        self.allowed_options = allowed_options
        all_options = {'xs':xs, 'sm':sm, 'md':md, 'lg':lg, 'xl':xl}
        self.device_options = {DeviceClass(device_class_name): option_value for (device_class_name, option_value) in all_options.items() if option_value}
        if not all([i in self.allowed_options for i in self.device_options.values()]):
            raise ProgrammerError('Illegal option. Allowed options: %s, got: %s' % (self.allowed_options, self.device_options))

    def __len__(self):
        return self.device_options.__len__()

    def add_css_classes(self, html_widget, prefix=None):
        classes = []
        for device_class, value in self.device_options.items():
            prefix_to_use = prefix or self.prefix
            css_class = device_class.as_combined_css_class([prefix_to_use] if prefix_to_use else [],
                                                           [six.text_type(value)] if value is not True else [])
            html_widget.append_class(css_class)
            classes.append(css_class)

        return classes


class ContentJustification(ResponsiveOption):
    """Specifies how elements should be spaced horizontally inside a parent.

    For each device size specify alignment as one of:
      * 'start': to align at the start.
      * 'end': to align at the end.
      * 'center': to align in the center of available space.
      * 'between': distribute all empty space evenly between items, but with no space before the first element or after the last element.
      * 'around': distribute all empty space evenly between items, including before the first element and after the last element.

    :keyword xs: Alignment to use if the device is extra small.
    :keyword sm: Alignment to use if the device is small.
    :keyword md: Alignment to use if the device is medium.
    :keyword lg: Alignment to use if the device is large.
    :keyword xl: Alignment to use if the device is extra large.

    .. versionadded: 4.0
    """
    def __init__(self, **kwargs):
        super(ContentJustification, self).__init__(['start', 'center', 'end', 'between', 'around'], prefix='justify-content', **kwargs)


class Alignment(ResponsiveOption):
    """Specifies how elements should be aligned vertically inside a parent.

    For each device size specify alignment as one of:
      * 'start': to align at the top.
      * 'end': to align at the bottom.
      * 'center': to align in the center of available space.
      * 'baseline': to align with the bottom of the first text line of neighbouring elements.
      * 'stretch': stretches to fill all vertical space.

    :keyword xs: Alignment to use if the device is extra small.
    :keyword sm: Alignment to use if the device is small.
    :keyword md: Alignment to use if the device is medium.
    :keyword lg: Alignment to use if the device is large.
    :keyword xl: Alignment to use if the device is extra large.

    .. versionadded: 4.0
    """
    def __init__(self, **kwargs):
        super(Alignment, self).__init__(['start', 'end', 'center', 'baseline', 'stretch'], **kwargs)

    def add_css_classes(self, html_widget, prefix=None):
        if isinstance(html_widget.layout, ColumnLayout):
            scope = 'items'
        else:
            scope = 'self'

        return super(Alignment, self).add_css_classes(html_widget, prefix=prefix or 'align-%s' % scope)


class ResponsiveSize(ResponsiveOption):
    """A size used for layouts that can adapt depending on how big the user device is.

    Sizes kwargs for each device class are given as integers that denote a number of 12ths
    of the size of the container of the element being sized. Eg: 6 would mean 6 12ths, or
    1/2 the size of the container.

    If you specify a size for a device class, that size will be used for all devices of that
    class or bigger.

    It is not necessary to specify a size for every device class. By default, if a device
    class is omitted, it is assumed to be sized as per the nearest specified smaller device
    class. If there is no smaller device class, a value of 12/12ths is assumed.

    :keyword xs: Size to use if the device is extra small.
    :keyword sm: Size to use if the device is small.
    :keyword md: Size to use if the device is medium.
    :keyword lg: Size to use if the device is large.
    :keyword xl: Size to use if the device is extra large.

    """
    def __init__(self, **sizes):
        super(ResponsiveSize, self).__init__(range(1, 13), **sizes)

    def calculated_size_for(self, device_class):
        classes_that_impact = [device_class]+device_class.all_smaller
        for possible_class in reversed(classes_that_impact):
            try:
                return self.device_options[possible_class]
            except KeyError:
                pass
        return 0

    @classmethod
    def wraps_for_some_device_class(cls, all_options):
        return any([cls.wraps_for(device_class, all_options)
                   for device_class in DeviceClass.all_classes()])

    @classmethod
    def wraps_for(cls, device_class, all_options):
        return (cls.sum_sizes_for(device_class, all_options)) > 12

    @classmethod
    def sum_sizes_for(cls, device_class, all_options):
        total = 0
        for options in all_options:
            total += options.size.calculated_size_for(device_class)
            total += options.offsets.calculated_size_for(device_class)
        return total


class ColumnOptions(object):
    """Various options to change how a column should be displayed.

    :param name: The name of the column.
    :keyword size: The :class:`ResponsiveSize` of the column.
    :keyword offsets: A :class:`ResponsiveSize` representing extra space before the column.
    :keyword vertical_align: An :class:`Alignment` stating how this column should be aligned vertically in its container.

    .. versionadded: 4.0

    """
    @arg_checks(size=IsInstance(ResponsiveSize, allow_none=True),
                offsets=IsInstance(ResponsiveSize, allow_none=True),
                vertical_align=IsInstance(Alignment, allow_none=True))
    def __init__(self, name, size=None, offsets=None, vertical_align=None):
        self.name = name
        self.size = size or ResponsiveSize(xs=True)
        self.offsets = offsets or ResponsiveSize()
        self.vertical_align = vertical_align or Alignment()


class ColumnLayout(Layout):
    """A Layout that divides an element into a number of columns.

    Each positional argument passed to the constructor defines a
    column. Columns are added to the element using this Layout in the
    order they are passed to the constructor. Columns can also be
    added to the Widget later, by calling
    :meth:`ColumnLayout.add_column`.

    Each such column-defining argument to the constructor is a tuple
    of which the first element is the column name, and the second an
    instance of :class:`ColumnOptions`. Besides the size of the column,
    other adjustments can be made via :class:`ColumnOptions`.

    You can also pass the column name only (no tuple) in which case
    a default :class:`ColumnOptions` will be used.

    If an element is divided into a number of columns whose current
    combined width is wider than 12/12ths, the overrun flows to make
    an additional row.

    It is customary, for example to specify smaller sizes (ito 12ths)
    for bigger devices where you want the columns to fit in next to each
    other, but use BIGGER sizes (such as 12/12ths) for the columns for
    smaller sized devices. This has the effect that what was displayed
    as columns next to each other on the bigger device is displayed
    as "stacked" cells on a smaller device.

    By default, the smallest device classes are sized 12/12ths.

    By default, if you specify smaller column sizes for larger devices,
    but omit specifying a size for extra small (xs) devices, columns for xs
    devices are full width and thus stacked.

    .. versionchanged:: 4.0
       Each item in column_definitions can be just a string (the column name) or a tuple mapping a name to a :class:`ColumnOptions`.

    """
    def __init__(self, *column_definitions):
        super(ColumnLayout, self).__init__()
        if not all([isinstance(column_definition, six.string_types+(ColumnOptions,)) for column_definition in column_definitions]):
            raise ProgrammerError('All column definitions are expected be either a ColumnOptions object of a column name, got %s' % six.text_type(column_definitions))
        self.added_column_definitions = []
        self.add_slots = False
        self.add_gutters = True
        self.alignment = Alignment()
        self.content_justification = ContentJustification()
        self.columns = OrderedDict()  #: A dictionary containing the added columns, keyed by column name.
        self.column_definitions = OrderedDict()
        for column_definition in column_definitions:
            if isinstance(column_definition, six.string_types):
                name, options = column_definition, ColumnOptions(column_definition)
            else:
                name, options = column_definition.name, column_definition
            self.column_definitions[name] = options

    def with_slots(self):
        """Returns a copy of this ColumnLayout which will additionally add a Slot inside each added column,
           named for that column.
        """
        copy_with_slots = copy.deepcopy(self)
        copy_with_slots.add_slots = True
        return copy_with_slots

    def without_gutters(self):
        """Returns a copy of this ColumnLayout which will not display whitespace between columns.

           .. versionadded:: 4.0
        """
        copy_without_gutters = copy.deepcopy(self)
        copy_without_gutters.add_gutters = False
        return copy_without_gutters

    @arg_checks(content_justification=IsInstance(ContentJustification))
    def with_justified_content(self, content_justification):
        """Returns a copy of this ColumnLayout with justification options set on it.

           .. versionadded:: 4.0
        """
        copy_with_content_justification = copy.deepcopy(self)
        copy_with_content_justification.content_justification = content_justification
        return copy_with_content_justification

    @arg_checks(vertical_alignment=IsInstance(Alignment))
    def with_vertical_alignment(self, vertical_alignment):
        """Returns a copy of this ColumnLayout with the column alignment options set on it.

           .. versionadded:: 4.0
        """
        copy_with_alignment = copy.deepcopy(self)
        copy_with_alignment.alignment = vertical_alignment
        return copy_with_alignment

    def customise_widget(self):
        for name, options in self.column_definitions.items():
            self.add_column(options.name, size=options.size, offsets=options.offsets, vertical_align=options.vertical_align)
        self.widget.append_class('row')
        if not self.add_gutters:
            self.widget.append_class('no-gutters')
        self.alignment.add_css_classes(self.widget)
        self.content_justification.add_css_classes(self.widget)

    def add_clearfix(self, column_options):
        clearfix = self.widget.add_child(Div(self.view))
        clearfix.append_class('clearfix')
        wrapping_classes = [device_class for device_class in DeviceClass.all_classes()
                            if ResponsiveSize.wraps_for(device_class, self.added_column_definitions + [column_options])]
        if wrapping_classes:
            device_class = wrapping_classes[0]
            if device_class.one_smaller:
                clearfix.append_class(device_class.one_smaller.as_combined_css_class(['hidden'], []))
        return clearfix

    def add_column(self, name, size=None, offsets=None, vertical_align=None, column_widget=None):
        """Called to add a column with given options.

        :param name: (See :class:`ColumnOptions`)
        :keyword size: (See :class:`ColumnOptions`)
        :keyword offsets: (See :class:`ColumnOptions`)
        :keyword vertical_align: (See :class:`ColumnOptions`)
        :keyword column_widget: If given, this Widget is added as the column instead of a Div (the default).

        .. versionchanged: 4.0
           Changed to create a named column with all possible options.
        """
        column_options = ColumnOptions(name, size=size, offsets=offsets, vertical_align=vertical_align)
        if ResponsiveSize.wraps_for_some_device_class(self.added_column_definitions+[column_options]):
            self.add_clearfix(column_options)

        column = self.widget.add_child(column_widget or Div(self.view))

        column_options.size.add_css_classes(column, prefix='col')
        column_options.offsets.add_css_classes(column, prefix='offset')
        column_options.vertical_align.add_css_classes(column)

        self.added_column_definitions.append(column_options)
        self.columns[column_options.name] = column
        column.append_class('column-%s' % column_options.name)
        if self.add_slots:
            column.add_child(Slot(self.view, column_options.name))
        return column




# Copyright 2016, 2017, 2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. versionadded:: 3.2

Client-side popups and related utilities.


"""
from __future__ import print_function, unicode_literals, absolute_import, division

import six
import json

from reahl.component.i18n import Catalogue

from reahl.web.bootstrap.ui import A
from reahl.web.bootstrap.forms import ButtonStyle, ButtonSize

_ = Catalogue('reahl-web')


class JsObject(object):
    def __init__(self, **attributes):
        self.attributes = attributes

    def as_html_snippet(self):
        return '{%s}' % (','.join(['%s: %s' % (name, self.translate_to_js(value)) 
                                   for name, value in self.attributes.items()]))

    def translate_to_js(self, value):
        if hasattr(value, 'as_html_snippet'):
            return value.as_html_snippet()
        return json.dumps(value)

    def __setitem__(self, key, value):
        self.attributes[key] = value

    def __getitem__(self, key):
        return self.attributes[key]


class JsFunction(object):
    def __init__(self, body_text='', *args):
        self.args = args
        self.body_text = body_text
        
    def as_html_snippet(self):
        return 'function(%s){%s}' % (','.join(self.args), self.body_text)


class CheckCheckboxScript(JsFunction):
    def __init__(self, checkbox):
        self.checkbox = checkbox
        selector = self.checkbox.jquery_selector
        body_text = '''$(%s).filter("input").attr("checked", true);$(%s).find("input").attr("checked", true);''' % \
                               (selector, selector)
        super(CheckCheckboxScript, self).__init__(body_text=body_text)




class PopupA(A):
    def __init__(self, view, target_bookmark, show_for_selector, dismiss_label=None, close_button=True,
                 center_vertically=False, css_id=None):
        super(PopupA, self).__init__(view, target_bookmark.href, target_bookmark.description, css_id=css_id)
        self.set_title(target_bookmark.description)
        self.title = target_bookmark.description
        self.append_class('reahl-bootstrappopupa')
        self.show_for_selector = show_for_selector
        self.buttons = JsObject()
        self.dismiss_label = dismiss_label
        self.center_vertically = center_vertically
        if dismiss_label is None:
            self.dismiss_label = _('Close')
        if close_button:
            self.add_js_button(_('Close'))

    def add_js_button(self, label, js_function=None, style=None, size=None):
        options = [ButtonStyle(style) if style else None, ButtonSize(size) if size else None]

        js_function = js_function or JsFunction()
        css_classes = [option.as_html_snippet() for option in options if option]
        self.buttons[label] = JsObject(function=js_function, css_classes=css_classes)
        return self.buttons[label]

    @property
    def jquery_selector(self):
        return '"a.reahl-bootstrappopupa[href=\'%s\']"' % self.href

    def get_js(self, context=None):
        selector = self.contextualise_selector(self.jquery_selector, context)
        options = JsObject(showForSelector=self.show_for_selector,
                           buttons=self.buttons,
                           title=self.title,
                           dismissLabel=self.dismiss_label,
                           centerVertically=self.center_vertically)
        return ['$(%s).bootstrappopupa(%s);' % (selector, options.as_html_snippet())]





# Copyright 2016, 2017, 2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.stubble import stubclass
from reahl.tofu import scenario, Fixture, expected, NoException
from reahl.tofu.pytestsupport import with_fixtures

from reahl.webdev.tools import XPath

import reahl.web_dev.widgets.test_table
from reahl.component.modelinterface import Field, IntegerField, exposed
from reahl.component.context import ExecutionContext
from reahl.web.bootstrap.ui import Div
from reahl.web.bootstrap.tables import Table, StaticColumn, TableLayout, DataTable

from reahl.web_dev.fixtures import WebFixture
from reahl.web_dev.widgets.test_table import TableFixture
from reahl.component_dev.test_i18n import LocaleContextStub


class LayoutScenarios(Fixture):
    @scenario
    def header_dark(self):
        self.layout_kwargs = dict(dark=True)
        self.expected_table_css_class = 'table table-dark'
        self.expected_main_css_class = None

    @scenario
    def border(self):
        self.layout_kwargs = dict(border='cells')
        self.expected_table_css_class = 'table table-bordered'
        self.expected_main_css_class = None

    @scenario
    def border(self):
        self.layout_kwargs = dict(border='borderless')
        self.expected_table_css_class = 'table table-borderless'
        self.expected_main_css_class = None

    @scenario
    def striped_rows(self):
        self.layout_kwargs = dict(striped=True)
        self.expected_table_css_class = 'table table-striped'
        self.expected_main_css_class = None

    @scenario
    def hovered_rows(self):
        self.layout_kwargs = dict(highlight_hovered=True)
        self.expected_table_css_class = 'table table-hover'
        self.expected_main_css_class = None

    @scenario
    def compacted_cells(self):
        self.layout_kwargs = dict(compact=True)
        self.expected_table_css_class = 'table table-sm'
        self.expected_main_css_class = None

    @scenario
    def responsive(self):
        self.layout_kwargs = dict(responsive='lg')
        self.expected_table_css_class = 'table'
        self.expected_main_css_class = 'table-responsive-lg'

    @scenario
    def responsive_for_device(self):
        self.layout_kwargs = dict(responsive=True)
        self.expected_table_css_class = 'table'
        self.expected_main_css_class = 'table-responsive'


@with_fixtures(WebFixture, LayoutScenarios)
def test_table_layout_options(web_fixture, layout_scenarios):
    """TableLayout uses Bootstrap to implement many table layout options."""

    layout = TableLayout(**layout_scenarios.layout_kwargs)
    Table(web_fixture.view).use_layout(layout)
    assert layout.widget.table.get_attribute('class') == layout_scenarios.expected_table_css_class
    if layout_scenarios.expected_main_css_class:
        assert layout.widget.html_representation.get_attribute('class') == layout_scenarios.expected_main_css_class


class LayoutHeaderThemeScenarios(WebFixture):

    @scenario
    def no_theme(self):
        self.theme = None

    @scenario
    def light(self):
        self.theme = 'light'
        self.expected_css_class = 'thead-light'

    @scenario
    def dark(self):
        self.theme = 'dark'
        self.expected_css_class = 'thead-dark'


@with_fixtures(WebFixture, LayoutHeaderThemeScenarios)
def test_table_layout_header_options(web_fixture, layout_scenarios):
    """TableLayout can style a table header row."""

    data = [DataItem(1, 'T'), DataItem(2, 'H'), DataItem(3, 'E')]

    column_definitions = [StaticColumn(Field(label='Row Number'), 'row'),
                          StaticColumn(Field(label='Alpha'), 'alpha')]

    layout = TableLayout(heading_theme=layout_scenarios.theme)
    table = Table(web_fixture.view).with_data(column_definitions, data)
    table.use_layout(layout)

    if layout_scenarios.theme is not None:
        assert layout.widget.thead.get_attribute('class') == '%s' % layout_scenarios.expected_css_class
    else:
        assert not layout.widget.thead.has_attribute('class')


class DataItem(reahl.web_dev.widgets.test_table.DataItem):
    @exposed
    def fields(self, fields):
        fields.row = IntegerField(label='Row', required=True, default=self.row)
        fields.alpha = Field(label='Alpha', required=True, default=self.alpha)


class DataTableFixture(TableFixture):
    items_per_page = 3
        
    def new_columns(self):
        return [StaticColumn(IntegerField(label='Row Number'), 'row', sort_key=lambda i: i.row),
                StaticColumn(Field(label='Alpha'), 'alpha', sort_key=lambda i: i.alpha),
                ]

    def new_data(self):
        alphas = 'THEQUICKBROWNFXJMPSVLAZYDG'
        return [DataItem(number, alpha) for number, alpha in enumerate(alphas, start=1)]
        #[(1, 'T'), (2, 'H'), (3, 'E'), (4, 'Q'),...(23, 'Z'), (24, 'Y'), (25, 'D'), (26, 'G')]

    def new_MainWidget(self):
        fixture = self
        class MainWidget(Div):
            def __init__(self, view):
                super(MainWidget, self).__init__(view)

                data_table = DataTable(view, fixture.columns, fixture.data,
                                       items_per_page=fixture.items_per_page,
                                       caption_text='All my friends',
                                       summary='Summary for screen reader',
                                       css_id='my_table_data')
                self.add_child(data_table)
        return MainWidget

    def new_wsgi_app(self):
        return self.web_fixture.new_wsgi_app(enable_js=False, child_factory=self.MainWidget.factory())

    def xpath_for_sort_link_for_column(self, column_number):
        return '(//table/thead/tr/th)[%s]/a' % column_number

    def does_column_have_sort_link(self, column_number):
        column_header = self.web_fixture.driver_browser.web_driver.find_element_by_xpath('(//table/thead/tr/th)[%s]' % (column_number))
        return len(column_header.find_elements_by_tag_name('a')) == 1
        
    def is_column_sorted(self, column_number, direction):
        header_link = self.web_fixture.driver_browser.web_driver.find_element_by_xpath('(//table/thead/tr/th)[%s]/a' % (column_number))
        if direction:
            expected_class = 'sorted-%s' % direction
            return expected_class in header_link.get_attribute('class').split(' ')
        else:
            return all([expected_class not in header_link.get_attribute('class').split(' ') 
                        for expected_class in ['sorted-ascending','sorted-descending']])


@with_fixtures(WebFixture, DataTableFixture)
def test_paging_through_data(web_fixture, data_table_fixture):
    """DataTable splits its items into different pages (between which a user can navigate), showing only the items of a particular page at a time."""

    web_fixture.reahl_server.set_app(data_table_fixture.wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    #click to last page
    browser.click(XPath.link_starting_with_text('→'))
    browser.click(XPath.link_with_text('9'))

    assert data_table_fixture.table_number_rows() == 2
    assert data_table_fixture.get_table_row(1) == ['25' ,'D']
    assert data_table_fixture.get_table_row(2) == ['26' ,'G']

    #click to page 4
    browser.click(XPath.link_starting_with_text('←'))
    browser.click(XPath.link_with_text('4'))

    assert data_table_fixture.table_number_rows() == 3
    assert data_table_fixture.get_table_row(1) == ['10' ,'R']
    assert data_table_fixture.get_table_row(2) == ['11' ,'O']
    assert data_table_fixture.get_table_row(3) == ['12' ,'W']


@with_fixtures(WebFixture, DataTableFixture)
def test_sorting(web_fixture, data_table_fixture):
    """By clicking on special links in the column header, the table is sorted according to that column - ascending or descending."""

    web_fixture.reahl_server.set_app(data_table_fixture.wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    #----- by default, not sorted
    assert not data_table_fixture.is_column_sorted(1, 'ascending')
    assert not data_table_fixture.is_column_sorted(1, 'descending')
    assert not data_table_fixture.is_column_sorted(2, 'ascending')
    assert not data_table_fixture.is_column_sorted(2, 'descending')

    #----- first click on column sorts ascending
    browser.click(data_table_fixture.xpath_for_sort_link_for_column(1))
    assert data_table_fixture.is_column_sorted(1, 'ascending')
    assert data_table_fixture.get_table_row(1) == ['1' ,'T']
    assert data_table_fixture.get_table_row(2) == ['2' ,'H']
    assert data_table_fixture.get_table_row(3) == ['3' ,'E']

    #----- sort ascending on alpha, the second column
    assert data_table_fixture.is_column_sorted(2, None)
    browser.click(data_table_fixture.xpath_for_sort_link_for_column(2))
    assert data_table_fixture.is_column_sorted(2, 'ascending')
    assert data_table_fixture.is_column_sorted(1, None)

    assert data_table_fixture.get_table_row(1) == ['22' ,'A']
    assert data_table_fixture.get_table_row(2) == ['9' ,'B']
    assert data_table_fixture.get_table_row(3) == ['7' ,'C']

    #----- sort descending on alpha, the second column
    browser.click(data_table_fixture.xpath_for_sort_link_for_column(2))
    assert data_table_fixture.is_column_sorted(2, 'descending')

    assert data_table_fixture.get_table_row(1) == ['23' ,'Z']
    assert data_table_fixture.get_table_row(2) == ['24' ,'Y']
    assert data_table_fixture.get_table_row(3) == ['15' ,'X']

    #----- sort order stays changed when paging
    browser.click(XPath.link_with_text('4'))

    assert data_table_fixture.get_table_row(1) == ['4' ,'Q']
    assert data_table_fixture.get_table_row(2) == ['18' ,'P']
    assert data_table_fixture.get_table_row(3) == ['11' ,'O']

    #----- contents of the page you are on changes according to a new sort order
    browser.click(data_table_fixture.xpath_for_sort_link_for_column(1))

    assert data_table_fixture.get_table_row(1) == ['10' ,'R']
    assert data_table_fixture.get_table_row(2) == ['11' ,'O']
    assert data_table_fixture.get_table_row(3) == ['12' ,'W']


@with_fixtures(WebFixture, DataTableFixture)
def test_sorting_i18n(web_fixture, data_table_fixture):
    """Sorting works correctly when the current View is translated to another i18n locale."""

    LocaleContextStub(locale='af').install()

    web_fixture.reahl_server.set_app(data_table_fixture.wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/af/')

    assert not data_table_fixture.is_column_sorted(1, 'ascending')
    assert not data_table_fixture.is_column_sorted(1, 'descending')
    browser.click(data_table_fixture.xpath_for_sort_link_for_column(1))
    assert data_table_fixture.is_column_sorted(1, 'ascending')


@with_fixtures(WebFixture, DataTableFixture)
def test_which_columns_can_cause_sorting(web_fixture, data_table_fixture):
    """Only columns with sort_key specified are sortable."""

    data_table_fixture.columns.append(StaticColumn(Field(label='Not sortable'), 'alpha'))

    web_fixture.reahl_server.set_app(data_table_fixture.wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    assert data_table_fixture.does_column_have_sort_link(1)
    assert data_table_fixture.does_column_have_sort_link(2)
    assert not data_table_fixture.does_column_have_sort_link(3)


@with_fixtures(WebFixture, DataTableFixture)
def test_layout_for_contained_table(web_fixture, data_table_fixture):
    """You can specify a Layout to use for the actual table inside the DataTable."""

    layout = TableLayout(heading_theme='light')  # heading_theme is here to give coverage of a previous bug
    data_table = DataTable(web_fixture.view, data_table_fixture.columns, data_table_fixture.data, 'my_css_id', table_layout=layout)

    assert data_table.table.layout is layout





# Copyright 2015-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. versionadded:: 3.2

Carousel presents a slideshow of Widgets with captions on each slide.


"""
from __future__ import print_function, unicode_literals, absolute_import, division

import six
from reahl.web.fw import Widget, Url
from reahl.web.ui import HTMLAttributeValueOption, HTMLElement, Img
from reahl.web.bootstrap.ui import Div, A, Span, Li, Ol, TextNode
from reahl.component.i18n import Catalogue

_ = Catalogue('reahl-web')


class Slide(Div):
    def __init__(self, view, widget, caption_widget, index):
        super(Slide, self).__init__(view)
        self.index = index
        self.append_class('carousel-item')
        self.add_child(widget)
        if self.is_active:
            self.append_class('active')

        if isinstance(widget, Img):
            widget.append_class('d-block')
            widget.append_class('w-100')

        if caption_widget:
            self.add_child(self.create_caption(caption_widget))

    @property
    def is_active(self):
        return self.index == 0

    def create_caption(self, caption_widget):
        div = Div(self.view)
        div.append_class('carousel-caption')
        div.add_child(caption_widget)
        return div


class Carousel(Widget):
    """
    A Widget that appears as having multiple panels of content of which
    only one panel is visible at a time. The visible panel is
    exchanged for another panel of content when prompted or
    automatically, at regular intervals, by visually sliding the new
    panel into position thus creating a slideshow effect.

    :param view: (See :class:`~reahl.web.fw.Widget`)
    :param css_id: (See :class:`~reahl.web.ui.HTMLElement`)
    :keyword show_indicators: If True (the default), includes indicators showing the number of slides and which one is visible.
    :keyword interval: The number (an int) of milliseconds to delay between cycling content. If not given (the default), the Carousel will not cycle automatically.
    :keyword pause: If set to 'false', hovering over the carousel will not pause it. Another option, 'hover' is supported.
    :keyword wrap: If True, the Carousel cycles contiuously, else it stops at the last slide.
    :keyword keyboard: If True, the Carousel reacts to keyboard events.
    :keyword min_height: If given, force slides to be at least this high (useful to prevent
             resizing of the Carousel between slides that are unequal in height). The value 
             is an int denoting a height in terms of the size of the font of the contents (em). 
    """
    def __init__(self, view, css_id, show_indicators=True, interval=5000, pause='hover', wrap=True, keyboard=True, min_height=None):
        super(Carousel, self).__init__(view)
        self.carousel_panel = self.add_child(Div(view, css_id=css_id))
        self.carousel_panel.append_class('carousel')
        self.carousel_panel.append_class('slide')
        self.carousel_panel.set_attribute('data-ride', 'carousel')

        self.carousel_panel.set_attribute('data-interval', six.text_type(interval))
        pause_option = HTMLAttributeValueOption(pause or 'false', True, constrain_value_to=['hover', 'false'])
        self.carousel_panel.set_attribute('data-pause', pause_option.as_html_snippet())
        self.carousel_panel.set_attribute('data-wrap', 'true' if wrap else 'false')
        self.carousel_panel.set_attribute('data-keyboard', 'true' if keyboard else 'false')
        if min_height:
            style = self.carousel_panel.add_child(HTMLElement(self.view, 'style', children_allowed=True))
            css_id = self.carousel_panel.css_id
            style.add_child(TextNode(self.view, '#%s .carousel-item { min-height: %sem; }' % (css_id, min_height)))

        self.show_indicators = show_indicators
        if self.show_indicators:
            self.indicator_list = self.carousel_panel.add_child(self.create_indicator_list())
        self.inner = self.carousel_panel.add_child(self.create_inner())
        self.slides = []

        self.add_control(previous=True)
        self.add_control()

    def create_inner(self):
        inner = Div(self.view)
        inner.append_class('carousel-inner')
        return inner

    def add_slide(self, widget, caption_widget=None):
        """
        Adds a Slide for the given :class:`~reahl.web.fw.Widget`.

        :param widget: A :class:`~reahl.web.fw.Widget` to display in this slide. (See also :class:`~reahl.web.holder.holder.PlaceholderImage`)
        :param caption_widget: The :class:`~reahl.web.fw.Widget` to use as caption for the slide.

        """
        slide = self.inner.add_child(Slide(self.view, widget, caption_widget, len(self.slides)))

        self.slides.append(slide)

        if self.show_indicators:
            self.add_indicator_for(slide)

        return slide

    @property
    def url(self):
        return Url('#%s' % self.carousel_panel.css_id)

    def add_control(self, previous=False):
        control_a = self.carousel_panel.add_child(A(self.view, self.url))
        control_a.append_class('carousel-control-prev' if previous else 'carousel-control-next')
        control_a.set_attribute('role', 'button')
        control_a.set_attribute('data-slide', 'prev' if previous else 'next')

        span_icon = control_a.add_child(Span(self.view))
        span_icon.append_class('carousel-control-%s-icon' % ('prev' if previous else 'next'))
        span_icon.set_attribute('aria-hidden', 'true')

        span_text = control_a.add_child(Span(self.view, text=_('Previous') if previous else _('Next')))
        span_text.append_class('sr-only')

        return control_a

    def create_indicator_list(self):
        indicators = Ol(self.view)
        indicators.append_class('carousel-indicators')
        return indicators

    def add_indicator_for(self, item):
        li = self.indicator_list.add_child(Li(self.view))
        li.set_attribute('data-target', six.text_type(self.url))
        li.set_attribute('data-slide-to', '%s' % item.index)
        if item.is_active:
            li.append_class('active')
        return li



# Copyright 2015-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.tofu import scenario, Fixture, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.webdev.tools import WidgetTester

from reahl.web.bootstrap.ui import P
from reahl.web.bootstrap.forms import TextInput, InputGroup

from reahl.web_dev.fixtures import WebFixture
from reahl.web_dev.inputandvalidation.test_input import SimpleInputFixture

@uses(web_fixture=WebFixture, simple_input_fixture=SimpleInputFixture)
class InputGroupFixture(Fixture):

    def new_an_input(self):
        return TextInput(self.simple_input_fixture.form, self.simple_input_fixture.field)

    @scenario
    def plain_text(self):
        self.input_group = InputGroup('before text', self.an_input, 'after text')
        self.expects_before_html = '<span class="input-group-prepend"><span class="input-group-text">before text</span></span>'
        self.expects_after_html = '<span class="input-group-append"><span class="input-group-text">after text</span></span>'

    @scenario
    def none_specified(self):
        self.input_group = InputGroup(None, self.an_input, None)
        self.expects_before_html = ''
        self.expects_after_html = ''

    @scenario
    def widgets(self):
        self.input_group = InputGroup(P(self.web_fixture.view, text='before widget'),
                                      self.an_input, 
                                      P(self.web_fixture.view, text='after widget'))
        self.expects_before_html = '<span class="input-group-prepend"><p class="input-group-text">before widget</p></span>'
        self.expects_after_html = '<span class="input-group-append"><p class="input-group-text">after widget</p></span>'


@with_fixtures(WebFixture, InputGroupFixture)
def test_input_group(web_fixture, input_group_fixture):
    """An InputGroup is a composition of an input with some text or Widget before and/or after an input."""
    fixture = input_group_fixture

    tester = WidgetTester(fixture.input_group)

    [outer_div] = tester.xpath('//div')
    assert outer_div.attrib['class'] == 'input-group'

    if fixture.expects_before_html:
        rendered_html = tester.get_html_for('//div/input/preceding-sibling::span')
        assert rendered_html == fixture.expects_before_html
    else:
        assert not tester.is_element_present('//div/input/preceding-sibling::span')

    children = outer_div.getchildren()
    the_input = children[1] if fixture.expects_before_html else children[0]
    assert the_input.tag == 'input'
    assert the_input.name == 'an_attribute'

    if fixture.expects_after_html:
        rendered_html = tester.get_html_for('//div/input/following-sibling::span')
        assert rendered_html == fixture.expects_after_html
    else:
        assert not tester.is_element_present('//div/input/following-sibling::span')




# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Configuration for reahl-web.
"""
from __future__ import print_function, unicode_literals, absolute_import, division

import os

from reahl.component.config import Configuration, ConfigSetting
from reahl.web.libraries import LibraryIndex, JQuery, JQueryUI, HTML5Shiv, IE9, Reahl, Holder, Popper, Bootstrap4, ReahlBootstrap4Additions


class WebConfig(Configuration):
    filename = 'web.config.py'
    config_key = 'web'

    site_root = ConfigSetting(description='The UserInterface class to be used as the root of the web application')
    static_root = ConfigSetting(default=os.getcwd(),
                                description='The directory from which static files will be served',
                                dangerous=True)
    frontend_libraries = ConfigSetting(description='A collection of front end libraries to include on pages')
    session_key_name = ConfigSetting(default='reahl',
                                     description='The name of this site\'s cookie in a user\'s a browser')
    guest_key_name = ConfigSetting(default='reahl-guest',
                                   description='The name of the cookie used to store information about anonymous visitors')
    guest_key_lifetime = ConfigSetting(default=60*60*24*365,
                                       description='The expiry time in seconds of the anonymous visitors cookie')

    session_class = ConfigSetting(description='The class used for UserSessions',
                                  automatic=True)
    persisted_exception_class = ConfigSetting(description='The class used for PersistedExceptions',
                                              automatic=True)
    persisted_userinput_class = ConfigSetting(description='The class used for UserInput',
                                              automatic=True)
    persisted_file_class = ConfigSetting(description='The class used for PersistedFiles',
                                              automatic=True)
    
    encrypted_http_scheme = ConfigSetting(default='https',
                                          description='The http scheme used for encrypted connections.')
    default_http_scheme = ConfigSetting(default='http',
                                        description='The http scheme used when an encrypted connection is not required' )
    encrypted_http_port = ConfigSetting(default='8363',
                                        description='The http port used for encrypted connections.',
                                        dangerous=True)
    default_http_port = ConfigSetting(default='8000',
                                      description='The http port used when an encrypted connection is not required',
                                      dangerous=True )
    default_url_locale = ConfigSetting(default='en_gb',
                                       description='The locale used when no locale is present in an URL')
    cache_max_age = ConfigSetting(default=10*60,
                                  description='The max time (in seconds) a cacheable dynamic page should be cached for')
    session_lifetime = ConfigSetting(default=60*60*24*7*2,
                                     description='The time in seconds a user session will be kept after last use')
    idle_lifetime = ConfigSetting(default=60*60*2,
                                  description='The time in seconds after which a user session will be considered idle - normal setting')
    idle_lifetime_max = ConfigSetting(default=60*60*24*7*2,
                                      description='The time in seconds after which a user session will be considered idle - "forever" setting')
    idle_secure_lifetime = ConfigSetting(default=60*60,
                                         description='The time in seconds after which a secure session will be considered expired')
                                       
    @property
    def secure_key_name(self):
        return '%s_secure' % self.session_key_name

    def __init__(self):
        super(WebConfig, self).__init__()
        # We create it here, so that each instance of a WebConfig will have its own LibraryIndex instance
        self.frontend_libraries = LibraryIndex(JQuery(), JQueryUI(), HTML5Shiv(), IE9(), Reahl(), Holder(),
                                               Popper(),  # must be before Bootstrap in html script includes
                                               Bootstrap4(), ReahlBootstrap4Additions())




# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division

from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import stubclass

from reahl.component.modelinterface import Field, ValidationConstraint, RequiredConstraint, MinLengthConstraint, \
                             MaxLengthConstraint, PatternConstraint, AllowedValuesConstraint, \
                             EqualToConstraint, RemoteConstraint, exposed
from reahl.web.ui import PrimitiveInput, HTMLInputElement, Form, TextInput, ButtonInput
from reahl.webdev.tools import WidgetTester

from reahl.web_dev.inputandvalidation.test_input import SimpleInputFixture2
from reahl.web_dev.fixtures import WebFixture


class ConstraintRenderingFixture(SimpleInputFixture2):
    def new_field(self, name='an_attribute', label='an attribute'):
        field = super(ConstraintRenderingFixture, self).new_field(label=label)
        field.bind(name, self.model_object)
        return field

    def new_error_xpath(self):
        return '//form/label[@class="error"]'

    def is_error_text(self, text):
        return text == self.web_fixture.driver_browser.get_text(self.error_xpath)

    def new_input(self, field=None):
        @stubclass(PrimitiveInput)
        class InputStub(PrimitiveInput):
            def create_html_widget(self):
                return HTMLInputElement(self, 'inputtype')

        return InputStub(self.form, field or self.field)


@with_fixtures(WebFixture, ConstraintRenderingFixture)
def test_rendering_of_constraints(web_fixture, constraint_rendering_fixture):
    """The constraints of the Field of an PrimitiveInput are rendered in html as html attributes of an Input
       which corresponds with the name of each validation_constraint and has value the parameters of the validation_constraint.
       The error message of each validation_constraint is also put in a json object inside the class attribute.
       These measures make it possible to write constraints that are checked on the browser either by
       browser support or by a jquery validate add-on, and in case of failure to display the exact
       corresponding error message."""

    fixture = constraint_rendering_fixture

    fixture.field.bind('an_attribute', fixture.model_object)
    fixture.model_object.an_attribute = 'field value'

    constraint1 = ValidationConstraint('validation_constraint 1 message')
    constraint1.name = 'one'
    @stubclass(ValidationConstraint)
    class ConstraintWithParams(ValidationConstraint):
        @property
        def parameters(self):
            return 'a parameter'

    constraint2 = ConstraintWithParams('validation_constraint 2 message with apostrophe\'s and quotes"')
    constraint2.name = 'two'

    fixture.field.add_validation_constraint(constraint1)
    fixture.field.add_validation_constraint(constraint2)

    tester = WidgetTester(fixture.input)

    actual = tester.render_html()
    expected_html = '''<input name="an_attribute" data-msg-one="validation_constraint 1 message" data-msg-two="validation_constraint 2 message with apostrophe&#x27;s and quotes&quot;" data-rule-one="true" data-rule-two="a parameter" form="test" type="inputtype" value="field value">'''
    assert actual == expected_html


@with_fixtures(WebFixture, ConstraintRenderingFixture)
def test_remote_constraints(web_fixture, constraint_rendering_fixture):
    """Remote constraints are invoked by the browser via ajax on the server when the input loses focus."""

    fixture = constraint_rendering_fixture

    class MyRemoteConstraint(RemoteConstraint):
        def validate_input(self, unparsed_input):
            if unparsed_input == 'failing_string_value':
                raise self

        def validate_parsed_value(self, parsed_value):
            if parsed_value == 'failing_parsed_value':
                raise self

    model_object = fixture.model_object
    model_object.fields.an_attribute.add_validation_constraint(MyRemoteConstraint(error_message='$label is invalid'))

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            field = model_object.fields.an_attribute
            self.add_child(TextInput(self, model_object.fields.an_attribute))
            self.define_event_handler(model_object.events.an_event)
            self.add_child(ButtonInput(self, model_object.events.an_event))

    wsgi_app = web_fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)

    web_fixture.driver_browser.open('/')

    # Initially, there's no error
    web_fixture.driver_browser.wait_for_element_not_visible(fixture.error_xpath)

    # A failing string value causes an ajax call resulting in an error
    web_fixture.driver_browser.type('//input[@type="text"]', 'failing_string_value')
    web_fixture.driver_browser.press_tab()
    web_fixture.driver_browser.wait_for_element_visible(fixture.error_xpath)
    assert fixture.is_error_text('an attribute is invalid')

    # A passing value causes an ajax call resulting in clearing of any previous errors
    web_fixture.driver_browser.type('//input[@type="text"]', 'passing value')

    web_fixture.driver_browser.press_tab()
    web_fixture.driver_browser.wait_for_element_not_visible(fixture.error_xpath)

    # A failing python value causes an ajax call resulting in an error
    web_fixture.driver_browser.type('//input[@type="text"]', 'failing_parsed_value')
    web_fixture.driver_browser.press_tab()
    web_fixture.driver_browser.wait_for_element_visible(fixture.error_xpath)
    assert fixture.is_error_text('an attribute is invalid')


@with_fixtures(WebFixture, ConstraintRenderingFixture)
def test_required_constraint_js(web_fixture, constraint_rendering_fixture):
    fixture = constraint_rendering_fixture

    constraint = RequiredConstraint()

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            field = fixture.model_object.fields.an_attribute.with_validation_constraint(constraint)
            self.add_child(TextInput(self, field))
    wsgi_app = web_fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)

    web_fixture.driver_browser.open('/')

    web_fixture.driver_browser.type('//input[@type="text"]', 'something')
    web_fixture.driver_browser.press_tab()
    web_fixture.driver_browser.wait_for_element_not_visible(fixture.error_xpath)

    web_fixture.driver_browser.type('//input[@type="text"]', '')
    web_fixture.driver_browser.press_backspace('//input')  # To trigger validation on the field

    web_fixture.driver_browser.wait_for_element_visible(fixture.error_xpath)


@with_fixtures(WebFixture, ConstraintRenderingFixture)
def test_min_length_constraint_js(web_fixture, constraint_rendering_fixture):
    fixture = constraint_rendering_fixture

    min_required_length = 5
    constraint = MinLengthConstraint(min_length=min_required_length)
    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            field = fixture.model_object.fields.an_attribute.with_validation_constraint(constraint)
            self.add_child(TextInput(self, field))

    wsgi_app = web_fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)

    web_fixture.driver_browser.open('/')

    web_fixture.driver_browser.type('//input[@type="text"]', '1234')
    web_fixture.driver_browser.press_tab()
    web_fixture.driver_browser.wait_for_element_visible(fixture.error_xpath)


@with_fixtures(WebFixture, ConstraintRenderingFixture)
def test_max_length_constraint_js(web_fixture, constraint_rendering_fixture):
    fixture = constraint_rendering_fixture

    max_allowed_length = 5
    constraint = MaxLengthConstraint(max_length=max_allowed_length)
    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            field = fixture.model_object.fields.an_attribute.with_validation_constraint(constraint)
            self.add_child(TextInput(self, field))

    wsgi_app = web_fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)

    web_fixture.driver_browser.open('/')

    web_fixture.driver_browser.type('//input[@type="text"]', '123456')
    accepted_value = web_fixture.driver_browser.get_value('//input[@type="text"]')
    assert accepted_value == '12345'


@with_fixtures(WebFixture, ConstraintRenderingFixture)
def test_pattern_constraint_js(web_fixture, constraint_rendering_fixture):
    fixture = constraint_rendering_fixture

    allow_pattern = '(ab)+'
    constraint = PatternConstraint(pattern=allow_pattern)

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            field = fixture.model_object.fields.an_attribute.with_validation_constraint(constraint)
            self.add_child(TextInput(self, field))

    wsgi_app = web_fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)

    web_fixture.driver_browser.open('/')

    web_fixture.driver_browser.type('//input[@type="text"]', 'aba')
    web_fixture.driver_browser.press_tab()
    web_fixture.driver_browser.wait_for_element_visible(fixture.error_xpath)

    web_fixture.driver_browser.type('//input[@type="text"]', 'ababab')
    web_fixture.driver_browser.press_tab()
    web_fixture.driver_browser.wait_for_element_not_visible(fixture.error_xpath)


@with_fixtures(WebFixture, ConstraintRenderingFixture)
def test_allowed_values_constraint_js(web_fixture, constraint_rendering_fixture):
    fixture = constraint_rendering_fixture

    allowed_values=['a','b']
    constraint = AllowedValuesConstraint(allowed_values=allowed_values)

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            field = fixture.model_object.fields.an_attribute.with_validation_constraint(constraint)
            self.add_child(TextInput(self, field))

    wsgi_app = web_fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)
    web_fixture.driver_browser.open('/')

    web_fixture.driver_browser.type('//input[@type="text"]', 'ba')
    web_fixture.driver_browser.press_tab()
    web_fixture.driver_browser.wait_for_element_visible(fixture.error_xpath)


@with_fixtures(WebFixture, ConstraintRenderingFixture)
def test_equal_to_constraint_js(web_fixture, constraint_rendering_fixture):
    fixture = constraint_rendering_fixture

    class ModelObject(object):
        @exposed
        def fields(self, fields):
            fields.an_attribute = Field(label='an attribute')
            fields.other = Field(label='other attribute')

    model_object = ModelObject()
    other_field = model_object.fields.other
    constraint = EqualToConstraint(other_field, '$label, $other_label')

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            field = fixture.model_object.fields.an_attribute.with_validation_constraint(constraint)
            other_input = self.add_child(TextInput(self, model_object.fields.other))
            other_input.set_id('other')
            one_input = self.add_child(TextInput(self, field))
            one_input.set_id('one')

    wsgi_app = web_fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)

    web_fixture.driver_browser.open('/')

    web_fixture.driver_browser.type('//input[@id="other"]', 'something')
    web_fixture.driver_browser.type('//input[@id="one"]', 'something else')
    web_fixture.driver_browser.press_tab()
    web_fixture.driver_browser.wait_for_element_visible(fixture.error_xpath)
    assert fixture.is_error_text('an attribute, other attribute')

    web_fixture.driver_browser.type('//input[@id="one"]', 'something')
    web_fixture.driver_browser.press_tab()
    web_fixture.driver_browser.wait_for_element_not_visible(fixture.error_xpath)



# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

Static pages that are included on the fly inside Views.
"""
from __future__ import print_function, unicode_literals, absolute_import, division

import os.path
import io
from six.moves import html_parser
import logging

from bs4 import BeautifulSoup, SoupStrainer

from reahl.component.modelinterface import Field
from reahl.component.exceptions import ProgrammerError
from reahl.component.i18n import Catalogue
from reahl.component.context import ExecutionContext
from reahl.web.fw import UrlBoundView, FileOnDisk, UserInterface, FileView, CannotCreate
from reahl.web.ui import LiteralHTML

_ = Catalogue('reahl-web')

class DJHTMLWidget(LiteralHTML):
    def __init__(self, view, html_content):
        super(DJHTMLWidget, self).__init__(view, html_content)

class DHTMLFile(object):
    def __init__(self, filename, ids):
        self.filename = filename
        self.ids = ids
        self.elements = {}
        self.title = None
        self.orignal_encoding = None
        
    def read(self):
        with io.open(self.filename, 'rb') as dhtml_file:
            def strain(name, attrs):
                if name == 'title':
                    return True
                if name == 'div' and dict(attrs).get('id', None) in self.ids:
                    return True
                return False
            soup = BeautifulSoup(dhtml_file, "lxml", parse_only=SoupStrainer(strain))
            parser = html_parser.HTMLParser()
            self.title = parser.unescape(soup.title.decode_contents()) if soup.title else _('Untitled')
            for an_id in self.ids:
                found_elements = soup.find_all(id=an_id)
                if found_elements:
                    number_of_ids = len(found_elements)
                    if number_of_ids != 1:
                        raise ProgrammerError('Expected to find one element with id "%s", but found %s' % (an_id, number_of_ids))
                    [element] = found_elements
                    self.elements[an_id] = element.decode_contents()
                else:
                    self.elements[an_id] = ''
            self.original_encoding = soup.original_encoding


class DhtmlUI(UserInterface):
    """A UserInterface which serves content from the static directory configured in `web.staticroot`.
       If a given Url maps directly to a file in this directory, that file is normally served
       as-is. If the filename ends on .d.html, however, the file is parsed, and the div inside it
       with id equal to `static_div_name` is read into the `main_slot` of a View. The
       title of the current View is also taken from the <title> of the static page.
       
       :param static_div_name: The id of the <div> to insert as `main_slot` of this UserInterface.
    """
    def assemble(self, static_div_name=None):
        self.static_div_name = static_div_name
        self.define_regex_view('(?P<file_path>.*)', '${file_path}', factory_method=self.create_view, file_path=Field())

    def is_dynamic(self, filename):
        return filename.endswith('.d.html') and os.path.exists(filename)

    def is_static(self, filename):
        return os.path.isfile(filename) and not filename.endswith('.d.html')

    def filesystem_path(self, relative_path):
        context = ExecutionContext.get_context()
        static_root = context.config.web.static_root
        if relative_path.endswith('/'):
           relative_path += 'index.d.html'
        return self.i18nise_filename(os.path.join(static_root, *relative_path.split('/')))

    def i18nise_filename(self, for_default_locale):
        current_locale = ExecutionContext.get_context().interface_locale
        head, tail = os.path.splitext(for_default_locale)
        head, d = os.path.splitext(head)
        for_current_locale = head+'.%s' % current_locale+d+tail
        if os.path.isfile(for_current_locale):
            return for_current_locale
        return for_default_locale

    def statics(self, relative_path):
        dhtml_file = DHTMLFile(self.filesystem_path(relative_path), [self.static_div_name])
        dhtml_file.read()
        statics = dict()
        statics['title'] = dhtml_file.title
        statics['div'] = dhtml_file.elements[self.static_div_name]
        return statics
    
    def create_view(self, relative_path, user_interface, file_path=None):
        if not user_interface is self:
            raise ProgrammerError('get_file called on %s with %s as user_interface' % (self, user_interface))
        file_url_path = file_path
        filename = self.filesystem_path(file_url_path)
        logging.debug('Finding a static file on filesystem %s' % filename)
        if self.is_dynamic(filename):
            statics = self.statics(file_url_path)
            view = UrlBoundView(user_interface, file_url_path, statics['title'], cacheable=True)
            view.set_slot('main_slot', DJHTMLWidget.factory(statics['div']))
            return view
        elif self.is_static(filename):
            return FileView(user_interface, FileOnDisk(filename, file_url_path))
        raise CannotCreate()


        




# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
import six
import json

from webob import Request

from reahl.tofu import Fixture, scenario, NoException, expected, uses
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import CallMonitor, EmptyStub

from reahl.webdev.tools import WidgetTester, XPath, Browser

from reahl.component.exceptions import DomainException, ProgrammerError, IsInstance
from reahl.component.modelinterface import IntegerField, EmailField, DateField, \
    exposed, Field, Event, Action, MultiChoiceField, Choice, AllowedValuesConstraint
from reahl.webdeclarative.webdeclarative import PersistedException, UserInput

from reahl.sqlalchemysupport import Session
from reahl.web.fw import Url, UserInterface, ValidationException
from reahl.web.ui import HTML5Page, Div, Form, TextInput, ButtonInput, NestedForm, SelectInput

from reahl.dev.fixtures import ReahlSystemFixture
from reahl.web_dev.fixtures import WebFixture, BasicPageLayout


@with_fixtures(WebFixture)
def test_basic_event_linkup(web_fixture):
    """When a user clicks on a Button, the Event to which the Button is linked is triggered on
       the server, and its corresponding action is executed. After this, the
       browser is transitioned to the target view as specified by the EventHandler.
    """
    fixture = web_fixture

    class ModelObject(object):
        handled_event = False
        def handle_event(self):
            self.handled_event = True

        @exposed
        def events(self, events):
            events.an_event = Event(label='click me', action=Action(self.handle_event))

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view, name, other_view):
            super(MyForm, self).__init__(view, name)
            self.define_event_handler(model_object.events.an_event, target=other_view)
            self.add_child(ButtonInput(self, model_object.events.an_event))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            home = self.define_view('/', title='Home page')
            other_view = self.define_view('/page2', title='Page 2')
            home.set_slot('main', MyForm.factory('myform', other_view))

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    fixture.reahl_server.set_app(wsgi_app)
    fixture.driver_browser.open('/')

    # clicking on the button triggers the action of the event handler
    assert not model_object.handled_event
    fixture.driver_browser.click("//input[@value='click me']")
    assert model_object.handled_event

    # browser has been transitioned to target view
    assert fixture.driver_browser.current_url.path == '/page2'


@with_fixtures(WebFixture)
def test_arguments_to_actions(web_fixture):
    """If a Button is created for an Event with_arguments, those arguments are passed to the backend
       when the Button is clicked."""

    fixture = web_fixture

    # how you link everything up in code
    class ModelObject(object):
        def handle_event(self, *args):
            self.args = args

        @exposed
        def events(self, events):
            events.an_event = Event(label='click me',
                                    action=Action(self.handle_event, ['one_argument', 'another_argument']),
                                    one_argument=IntegerField(),
                                    another_argument=Field())

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.add_child(ButtonInput(self, model_object.events.an_event.with_arguments(one_argument=1, another_argument='another')))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            home = self.define_view('/', title='Home page')
            other_view = self.define_view('/page2', title='Page 2')
            home.set_slot('main', MyForm.factory('myform'))
            self.define_transition(model_object.events.an_event, home, other_view)

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    fixture.reahl_server.set_app(wsgi_app)
    fixture.driver_browser.open('/')

    # when the Action is executed, the correct arguments are passed
    fixture.driver_browser.click("//input[@value='click me']")
    assert model_object.args == (1, 'another')


@with_fixtures(WebFixture)
def test_validation_of_event_arguments(web_fixture):
    """Buttons cannot be created for Events with invalid default arguments."""

    fixture = web_fixture

    class ModelObject(object):
        @exposed
        def events(self, events):
            events.an_event = Event(label='Click me', argument=Field(required=True))

    model_object = ModelObject()

    form = Form(fixture.view, 'test')
    form.define_event_handler(model_object.events.an_event)

    with expected(ProgrammerError):
        ButtonInput(form, model_object.events.an_event)

    with expected(NoException):
        ButtonInput(form, model_object.events.an_event.with_arguments(argument='something'))


@with_fixtures(WebFixture)
def test_basic_field_linkup(web_fixture):
    """When a form is rendered, the Inputs on it display either the default value as specified
       by the Field with which the Input is associated, or (if set) the value of the attribute
       that Field is bound to.

       When a user clicks on a Button, the Input value is set via the Field on the
       model objects to which the Field is bound. This happens before the associated action
       is executed.
    """

    fixture = web_fixture

    class ModelObject(object):
        def handle_event(self):
            pass

        @exposed
        def events(self, events):
            events.an_event = Event(label='click me', action=Action(self.handle_event))

        @exposed
        def fields(self, fields):
            fields.field_name = IntegerField(default=3)

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.define_event_handler(model_object.events.an_event)
            self.add_child(ButtonInput(self, model_object.events.an_event))
            self.add_child(TextInput(self, model_object.fields.field_name))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            home = self.define_view('/', title='Home page')
            home.set_slot('main', MyForm.factory('myform'))

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    fixture.reahl_server.set_app(wsgi_app)
    fixture.driver_browser.open('/')

    # the initial value rendered in the input
    assert not hasattr(model_object, 'field_name')
    initial_value = fixture.driver_browser.get_value("//input[@type='text']")
    assert initial_value == '3'

    # the user supplied value is marshalled and set on the model object when an event happens
    fixture.driver_browser.type("//input[@type='text']", '5')
    fixture.driver_browser.click("//input[@value='click me']")
    assert model_object.field_name == 5


@with_fixtures(WebFixture)
def test_distinguishing_identical_field_names(web_fixture):
    """A programmer can add different Inputs on the same Form even if their respective Fields are bound
       to identically named attributes of different objects."""

    fixture = web_fixture

    class ModelObject(object):
        @exposed
        def fields(self, fields):
            fields.field_name = IntegerField()

    model_object1 = ModelObject()
    model_object2 = ModelObject()

    class MyForm(Form):
        @exposed
        def events(self, events):
            events.an_event = Event(label='click me')

        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.define_event_handler(self.events.an_event)
            self.add_child(ButtonInput(self, self.events.an_event))

            self.add_child(TextInput(self, model_object1.fields.field_name))
            self.add_child(TextInput(self, model_object2.fields.field_name))

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('form'))
    fixture.reahl_server.set_app(wsgi_app)
    fixture.driver_browser.open('/')

    # the correct input value gets to the correct object despite referencing identically named attributes
    fixture.driver_browser.type('//input[@type="text"][1]', '0')
    fixture.driver_browser.type('//input[@type="text"][2]', '1')
    fixture.driver_browser.click(XPath.button_labelled('click me'))
    assert model_object1.field_name == 0
    assert model_object2.field_name == 1


@with_fixtures(WebFixture)
def test_wrong_arguments_to_define_event_handler(web_fixture):
    """Passing anything other than an Event to define_event_handler is an error."""

    fixture = web_fixture

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.define_event_handler(EmptyStub())

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('form'))
    browser = Browser(wsgi_app)

    with expected(IsInstance):
        browser.open('/')


@with_fixtures(WebFixture)
def test_define_event_handler_not_called(web_fixture):
    """."""
    fixture = web_fixture

    class ModelObject(object):
        @exposed
        def events(self, events):
            events.an_event = Event()

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.add_child(ButtonInput(self, model_object.events.an_event))

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('form'))
    browser = Browser(wsgi_app)

    with expected(ProgrammerError, test='no Event/Transition available for name an_event'):
        browser.open('/')


@with_fixtures(ReahlSystemFixture, WebFixture)
def test_exception_handling(reahl_system_fixture, web_fixture):
    """When a DomainException happens during the handling of an Event:

       The database is rolled back.
       The browser is redirected to GET the original view again (not the target).
       The screen still displays the values the user initially typed, not those on the ModelObject.
    """

    fixture = web_fixture

    class ModelObject(object):
        def handle_event(self):
            self.field_name = 1
            raise DomainException()
        @exposed
        def events(self, events):
            events.an_event = Event(label='click me', action=Action(self.handle_event))
        @exposed
        def fields(self, fields):
            fields.field_name = IntegerField(default=3)

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view, name, other_view):
            super(MyForm, self).__init__(view, name)
            self.define_event_handler(model_object.events.an_event, target=other_view)
            self.add_child(ButtonInput(self, model_object.events.an_event))
            self.add_child(TextInput(self, model_object.fields.field_name))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            home = self.define_view('/', title='Home page')
            other_view = self.define_view('/page2', title='Page 2')
            home.set_slot('main', MyForm.factory('myform', other_view))

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    fixture.reahl_server.set_app(wsgi_app)

    fixture.driver_browser.open('/')

    assert not hasattr(model_object, 'field_name')
    fixture.driver_browser.type("//input[@type='text']", '5')

    # any database stuff that happened when the form was submitted was rolled back
    with CallMonitor(reahl_system_fixture.system_control.orm_control.rollback) as monitor:
        fixture.driver_browser.click(XPath.button_labelled('click me'))
    assert monitor.times_called == 1

    # the value input by the user is still displayed on the form, NOT the actual value on the model object
    assert model_object.field_name == 1
    retained_value = fixture.driver_browser.get_value("//input[@type='text']")
    assert retained_value == '5'

    # the browser is still on the page with the form which triggered the exception
    assert fixture.driver_browser.current_url.path == '/'


@with_fixtures(WebFixture)
def test_form_preserves_user_input_after_validation_exceptions_multichoice(web_fixture):
    """When a form is submitted and validation fails on the server, the user
     is presented with the values that were originally entered (even if they were invalid)."""

    fixture = web_fixture

    class ModelObject(object):
        @exposed
        def events(self, events):
            events.an_event = Event(label='click me')
        @exposed
        def fields(self, fields):
            choices = [Choice(1, IntegerField(label='One')),
                       Choice(2, IntegerField(label='Two')),
                       Choice(3, IntegerField(label='Three'))]
            fields.no_validation_exception_field = MultiChoiceField(choices, label='Make your invalid choice', default=[])
            fields.validation_exception_field = MultiChoiceField(choices, label='Make your choice', default=[], required=True)

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view):
            super(MyForm, self).__init__(view, 'my_form')
            self.define_event_handler(model_object.events.an_event)
            self.add_child(ButtonInput(self, model_object.events.an_event))
            input = self.add_child(SelectInput(self, model_object.fields.no_validation_exception_field))
            if input.validation_error:
                self.add_child(self.create_error_label(input))
            input = self.add_child(SelectInput(self, model_object.fields.validation_exception_field))
            if input.validation_error:
                self.add_child(self.create_error_label(input))

    wsgi_app = web_fixture.new_wsgi_app(child_factory=MyForm.factory())
    browser = Browser(wsgi_app)

    browser.open('/')

    no_validation_exception_input = '//select[@name="no_validation_exception_field"]'
    validation_exception_input = '//select[@name="validation_exception_field"]'

    browser.select_many(no_validation_exception_input, ['One', 'Two'])
    browser.select_none(validation_exception_input) # select none to trigger the RequiredConstraint
    browser.click(XPath.button_labelled('click me'))

    assert browser.get_value(no_validation_exception_input) == ['1', '2']
    assert not browser.get_value(validation_exception_input)

    label = browser.get_html_for('//label')
    input_id = browser.get_id_of(validation_exception_input)
    assert label == '<label for="%s" class="error">Make your choice is required</label>' % input_id

    #2. Submit again ths time not expecting validation exceptions, also expecting the validation error to be cleared and the domain should have all input
    browser.select_many(validation_exception_input, ['Two', 'Three'])
    browser.click(XPath.button_labelled('click me'))

    assert not browser.is_element_present('//label[@class="error"]')
    assert browser.get_value(no_validation_exception_input) == ['1', '2']
    assert browser.get_value(validation_exception_input) == ['2', '3']


@with_fixtures(WebFixture)
def test_rendering_of_form(web_fixture):
    """A Form is always set up to POST to its EventChannel url.  The current page's query string is
       propagated with the POST url if any.  The Form has an id and class to help style it etc."""

    fixture = web_fixture

    form = Form(fixture.view, 'test_channel')
    tester = WidgetTester(form)

    fixture.context.request = Request.blank('/a/b?x=y', charset='utf8')
    actual = tester.render_html()

    expected = '<form id="test_channel" action="/a/b/_test_channel_method?x=y" data-formatter="/__test_channel_format_method" method="POST" class="reahl-form"></form>'
    assert actual == expected

    # Case: without querystring
    fixture.context.request = Request.blank('/a/b', charset='utf8')
    actual = tester.render_html_tree()

    action = actual.xpath('//form')[0].attrib['action']
    assert action == '/a/b/_test_channel_method'


@with_fixtures(WebFixture)
def test_duplicate_forms(web_fixture):
    """It is an error to add more than one form with the same unique_name to a page."""

    fixture = web_fixture

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout(slots=['main', 'secondary']))
            home = self.define_view('/', title='Home page')
            home.set_slot('main', Form.factory('myform'))
            home.set_slot('secondary', Form.factory('myform'))

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    with expected(ProgrammerError, test='More than one form was added using the same unique_name.*'):
        browser.open('/')


@with_fixtures(WebFixture)
def test_check_input_placement(web_fixture):
    """When a web request is handled, the framework throws an exception if an input might be seperated conceptually from the form they are bound to."""

    fixture = web_fixture

    class ModelObject(object):
        @exposed
        def fields(self, fields):
            fields.name = Field()

    class MainUI(UserInterface):
        def assemble(self):
            page = self.define_page(HTML5Page).use_layout(BasicPageLayout())
            home = self.define_view('/', title='Home page')
            home.set_slot('main', MyForm.factory())

    class MyForm(Form):
        def __init__(self, view):
            super(MyForm, self).__init__(view, 'my_form')
            self.add_child(RerenderableInputPanel(view, self))

    class RerenderableInputPanel(Div):
        def __init__(self, view, form):
            super(RerenderableInputPanel, self).__init__(view, css_id='my_refresh_id')
            self.enable_refresh()
            model_object = ModelObject()
            self.add_child(TextInput(form, model_object.fields.name))

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    with expected(ProgrammerError, test='.*Inputs are not allowed where they can be refreshed separately from their forms\..*'):
        browser.open('/')


@with_fixtures(WebFixture)
def test_check_missing_form(web_fixture):
    """All forms referred to by inputs on a page have to be present on that page."""

    fixture = web_fixture

    class ModelObject(object):
        @exposed
        def fields(self, fields):
            fields.name = Field()

    class MyPanel(Div):
        def __init__(self, view):
            super(MyPanel, self).__init__(view)
            model_object = ModelObject()
            forgotten_form = Form(view, 'myform')
            self.add_child(TextInput(forgotten_form, model_object.fields.name))

    wsgi_app = fixture.new_wsgi_app(child_factory=MyPanel.factory())
    browser = Browser(wsgi_app)

    expected_message = 'Could not find form for <TextInput name=name>. '\
                       'Its form, <Form form id=myform> is not present on the current page'

    with expected(ProgrammerError, test=expected_message):
        browser.open('/')


@with_fixtures(WebFixture)
def test_nested_forms(web_fixture):
    """HTML disallows nesting of forms. A NestedForm can be used to simulate
       a form which is visually part of another Form. A NestedForm provides a
       form attribute which should be used when creating an Inputs for it.
    """

    fixture = web_fixture

    class NestedModelObject(object):
        handled_event = False
        def handle_event(self):
            self.handled_event = True
        @exposed
        def events(self, events):
            events.nested_event = Event(label='click nested', action=Action(self.handle_event))
        @exposed
        def fields(self, fields):
            fields.nested_field = Field(label='input nested')

    nested_model_object = NestedModelObject()
    class MyNestedForm(NestedForm):
        def __init__(self, view, name):
            super(MyNestedForm, self).__init__(view, name)
            self.define_event_handler(nested_model_object.events.nested_event)
            self.add_child(ButtonInput(self.form, nested_model_object.events.nested_event))
            self.add_child(TextInput(self.form, nested_model_object.fields.nested_field))

    class OuterModelObject(object):
        handled_event = False
        def handle_event(self):
            self.handled_event = True
        @exposed
        def events(self, events):
            events.outer_event = Event(label='click outer', action=Action(self.handle_event))
    outer_model_object = OuterModelObject()
    class OuterForm(Form):
        def __init__(self, view, name):
            super(OuterForm, self).__init__(view, name)
            self.add_child(MyNestedForm(view, 'my_nested_form'))
            self.define_event_handler(outer_model_object.events.outer_event)
            self.add_child(ButtonInput(self, outer_model_object.events.outer_event))

    wsgi_app = fixture.new_wsgi_app(child_factory=OuterForm.factory('outer_form'))
    fixture.reahl_server.set_app(wsgi_app)
    browser = fixture.driver_browser

    browser.open('/')
    browser.type(XPath.input_named('nested_field'), 'some nested input')

    browser.click(XPath.button_labelled('click nested'))

    assert nested_model_object.handled_event
    assert not outer_model_object.handled_event

    assert nested_model_object.nested_field == 'some nested input'


@with_fixtures(WebFixture)
def test_form_input_validation(web_fixture):
    """Validation of input happens in JS on the client, but also on the server if JS is bypassed."""

    error_xpath = '//form[contains(@class, "reahl-form")]/label[contains(@class, "error")]'

    fixture = web_fixture

    class ModelObject(object):
        def handle_event(self):
            pass
        @exposed
        def events(self, events):
            events.an_event = Event(label='click me', action=Action(self.handle_event))
        @exposed
        def fields(self, fields):
            fields.field_name = EmailField()

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view, name, other_view):
            super(MyForm, self).__init__(view, name)
            self.define_event_handler(model_object.events.an_event, target=other_view)
            self.add_child(ButtonInput(self, model_object.events.an_event))
            text_input = self.add_child(TextInput(self, model_object.fields.field_name))
            if text_input.validation_error:
                self.add_child(self.create_error_label(text_input))


    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            home = self.define_view('/', title='Home page')
            other_view = self.define_view('/page2', title='Page 2')
            home.set_slot('main', MyForm.factory('myform', other_view))

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI, enable_js=True)

    # Case: form validation fails in JS on the client
    #  - Form submission is blocked
    #  - Error message is displayed
    fixture.reahl_server.set_app(wsgi_app)
    fixture.driver_browser.open('/')
    fixture.driver_browser.wait_for_element_not_visible(error_xpath)
    fixture.driver_browser.type('//input[@type="text"]', 'not@notvalid')
    fixture.driver_browser.press_tab()
    fixture.driver_browser.wait_for_element_visible(error_xpath)

    with fixture.driver_browser.no_page_load_expected():
        fixture.driver_browser.click("//input[@value='click me']")

    error_text = fixture.driver_browser.get_text(error_xpath)
    assert error_text == 'field_name should be a valid email address'

    # Case: form validation fails on the server (assuming no JS on the client to block submission)
    #  - ValidationException is raised (which is dealt with as any DomainException)
    browser = Browser(wsgi_app)
    browser.open('/')
    browser.type('//input[@type="text"]', 'not@notvalid')

    browser.click('//input[@type="submit"]')
    assert not hasattr(model_object, 'field_name')
    label = browser.get_html_for('//label')
    input_id = browser.get_id_of('//input[@type="text"]')
    assert label == '<label for="%s" class="error">field_name should be a valid email address</label>' % input_id

    assert Session.query(UserInput).filter_by(key='field_name').count() == 1  # The invalid input was persisted
    exception = Session.query(PersistedException).one().exception
    assert isinstance(exception, ValidationException)  # Is was persisted
    assert not exception.commit

    # Case: form validation passes (no-js)
    #  - no ValidationException
    #  - all input is translated to python and set as values on the model objects
    #  - any saved input on the form is cleared
    browser.type('//input[@type="text"]', 'valid@home.org')
    browser.click("//input[@value='click me']")
    assert model_object.field_name == 'valid@home.org'

    assert Session.query(UserInput).filter_by(key='field_name').count() == 0  # The invalid input was removed
    assert Session.query(PersistedException).count() == 0  # The exception was removed

    assert browser.location_path == '/page2'

    # Case: form validation passes (js)
    #  - no ValidationException
    #  - all input is translated to python and set as values on the model objects
    fixture.driver_browser.type('//input[@type="text"]', 'valid@home.org')
    fixture.driver_browser.wait_for_element_not_visible(error_xpath)
    fixture.driver_browser.click("//input[@value='click me']")
    assert model_object.field_name == 'valid@home.org'

    assert fixture.driver_browser.current_url.path == '/page2'


@uses(web_fixture=WebFixture)
class QueryStringScenarios(Fixture):
    def new_other_view(self):
        return self.web_fixture.wsgi_app.define_view('/page2', title='Page 2')

    def new_form(self):
        form = Form(self.web_fixture.view, 'some_form')
        event = Event(label='click me', action=Action(self.action))
        event.bind('an_event', None)
        form.define_event_handler(event, target=self.target)
        form.add_child(ButtonInput(form, event))
        return form

    @property
    def query_string_on_form_submit(self):
        form_action = self.web_fixture.driver_browser.get_attribute('//form', 'action')
        return Url(form_action).query

    @scenario
    def different_target(self):
        self.break_on_submit = False
        self.target_is_other_view = True
        self.initial_qs = 'a=b'
        self.final_qs = ''

    @scenario
    def same_target_by_intent(self):
        self.break_on_submit = False
        self.target_is_other_view = False
        self.target = self.web_fixture.view.as_factory()
        self.initial_qs = 'a=b'
        self.final_qs = 'a=b'

    @scenario
    def same_target_by_exception(self):
        self.break_on_submit = True
        self.target_is_other_view = True
        self.initial_qs = 'a=b'
        self.final_qs = 'a=b'


@with_fixtures(WebFixture, QueryStringScenarios)
def test_propagation_of_querystring(web_fixture, query_string_scenarios):
    """The query string of the original View visited is maintained through the GET/POST cycle.
       If, for whatever reason (exception or intent) the browser should stay on the same
       View after POSTing, the query string is maintained to that point.
       Else, the query string is cleared.
    """
    fixture = query_string_scenarios

    class ModelObject(object):
        def handle_event(self):
            if fixture.break_on_submit:
                raise DomainException()
        @exposed
        def events(self, events):
            events.an_event = Event(label='click me', action=Action(self.handle_event))

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view, name, other_view):
            super(MyForm, self).__init__(view, name)
            if fixture.target_is_other_view:
                target = other_view
            else:
                target = view.as_factory()
            self.define_event_handler(model_object.events.an_event, target=target)
            self.add_child(ButtonInput(self, model_object.events.an_event))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            home = self.define_view('/', title='Home page')
            other_view = self.define_view('/page2', title='Page 2')
            home.set_slot('main', MyForm.factory('myform', other_view))

    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    web_fixture.reahl_server.set_app(wsgi_app)
    web_fixture.driver_browser.open('/?%s' % fixture.initial_qs)

    assert fixture.query_string_on_form_submit == fixture.initial_qs
    web_fixture.driver_browser.click("//input[@value='click me']")
    assert web_fixture.driver_browser.current_url.query == fixture.final_qs


@with_fixtures(WebFixture)
def test_event_names_are_canonicalised(web_fixture):
    """The name= attribute of a button is an url encoded string. There is more than one way
       to url encode the same string. The server ensures that different encodings of the same
       string are not mistaken for different names.
    """

    fixture = web_fixture

    class ModelObject(object):
        def handle_event(self, some_argument):
            self.received_argument = some_argument

        @exposed
        def events(self, events):
            events.an_event = Event(label='click me',
                                    action=Action(self.handle_event, ['some_argument']),
                                    some_argument=Field(default='default value'))

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.define_event_handler(model_object.events.an_event)
            self.add_child(ButtonInput(self, model_object.events.an_event.with_arguments(some_argument='f~nnystuff')))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            home = self.define_view('/', title='Home page')
            home.set_slot('main', MyForm.factory('myform'))

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    # when the Action is executed, the correct arguments are passed
    browser.post('/__myform_method', {'event.an_event?some_argument=f~nnystuff': ''})
    assert model_object.received_argument == 'f~nnystuff'


@with_fixtures(WebFixture)
def test_alternative_event_trigerring(web_fixture):
    """Events can also be triggered by submitting a Form via Ajax. In such cases the normal redirect-after-submit
       behaviour of the underlying EventChannel is not desirable. This behaviour can be switched off by submitting
       an extra argument along with the Form in order to request the alternative behaviour.
    """

    fixture = web_fixture

    class ModelObject(object):
        def handle_event(self):
            self.handled_event = True

        @exposed
        def events(self, events):
            events.an_event = Event(label='click me',
                                    action=Action(self.handle_event))

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view, name, other_view):
            super(MyForm, self).__init__(view, name)
            self.define_event_handler(model_object.events.an_event, target=other_view)
            self.add_child(ButtonInput(self, model_object.events.an_event))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            home = self.define_view('/', title='Home page')
            other_view = self.define_view('/page2', title='Page 2')
            home.set_slot('main', MyForm.factory('myform', other_view))

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    # when POSTing with _noredirect, the Action is executed, but the browser is not redirected to /page2 as usual
    browser.post('/__myform_method', {'event.an_event?': '', '_noredirect': ''})
    browser.follow_response()  # Needed to make the test break should a HTTPTemporaryRedirect response be sent
    assert model_object.handled_event
    assert browser.location_path != '/page2'
    assert browser.location_path == '/__myform_method'

    # the response is a json object reporting the success of the event and a new rendition of the form
    json_dict = json.loads(browser.raw_html)
    assert json_dict['success']

    browser.open('/')
    expected_html = browser.get_inner_html_for('//form[1]')
    assert json_dict['widget'].startswith(expected_html+'<script')


@with_fixtures(WebFixture)
def test_remote_field_validation(web_fixture):
    """A Form contains a RemoteMethod that can be used to validate any of its fields via HTTP.
    """

    fixture = web_fixture

    class ModelObject(object):
        @exposed
        def fields(self, fields):
            fields.a_field = EmailField()

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.add_child(TextInput(self, model_object.fields.a_field))

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('myform'))
    browser = Browser(wsgi_app)

    browser.open('/_myform_validate_method?a_field=invalid email address')
    assert browser.raw_html == '"a_field should be a valid email address"'

    browser.open('/_myform_validate_method?a_field=valid@email.org')
    assert browser.raw_html == 'true'


@with_fixtures(WebFixture)
def test_remote_field_formatting(web_fixture):
    """A Form contains a RemoteMethod that can be used to reformat any of its fields via HTTP.
    """

    fixture = web_fixture

    class ModelObject(object):
        @exposed
        def fields(self, fields):
            fields.a_field = DateField()

    model_object = ModelObject()

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.add_child(TextInput(self, model_object.fields.a_field))

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('myform'))
    browser = Browser(wsgi_app)

    browser.open('/_myform_format_method?a_field=13 November 2012')
    assert browser.raw_html == '13 Nov 2012'

    browser.open('/_myform_format_method?a_field=invaliddate')
    assert browser.raw_html == ''



# Copyright 2016, 2017, 2018 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

""".. versionadded:: 3.2

A Bootstrap Navbar is a header for a web application with all manner
of useful content, including a navigation menu.

"""
from __future__ import print_function, unicode_literals, absolute_import, division

# noinspection PyUnresolvedReferences
import six

from reahl.component.exceptions import arg_checks, IsInstance, ProgrammerError

from reahl.web.fw import Layout
from reahl.web.ui import Url, Widget, HTMLWidget, HTMLElement, HTMLAttributeValueOption
from reahl.component.i18n import Catalogue
from reahl.web.bootstrap.ui import Div, Nav, A, TextNode, Span
from reahl.web.bootstrap.forms import Form
import reahl.web.bootstrap.navs
from reahl.web.bootstrap.grid import Container, DeviceClass

_ = Catalogue('reahl-web')


class CollapseToggle(HTMLElement):
    def __init__(self, view, target_widget, text=None):
        super(CollapseToggle, self).__init__(view, 'button', children_allowed=True)
        self.set_attribute('type', 'button')

        self.target_widget = target_widget
        self.append_class('navbar-toggler')
        self.set_attribute('data-toggle', 'collapse')
        self.set_attribute('data-target', '#%s' % target_widget.css_id)
        self.set_attribute('aria-controls', '%s' % target_widget.css_id)
        self.set_attribute('aria-label', _('Toggle navigation'))

        if text is None:
            collapse_toggle_widget = Span(self.view)
            collapse_toggle_widget.append_class('navbar-toggler-icon')
        else:
            collapse_toggle_widget = TextNode(view, text)
        self.add_child(collapse_toggle_widget)


class NavbarFixed(HTMLAttributeValueOption):
    def __init__(self, fixed_to):
        super(NavbarFixed, self).__init__(fixed_to, fixed_to is not None, 
                                          prefix='', constrain_value_to=['fixed-top', 'fixed-bottom', 'sticky-top'])


class NavbarTogglerAlignment(HTMLAttributeValueOption):
    def __init__(self, alignment):
        super(NavbarTogglerAlignment, self).__init__(alignment, alignment is not None,
                                                     prefix='navbar-toggler', constrain_value_to=['left', 'right'])


class ColourTheme(HTMLAttributeValueOption):
    def __init__(self, name):
        super(ColourTheme, self).__init__(name, name is not None, 
                                          prefix='navbar', constrain_value_to=['light', 'dark'])


class BackgroundScheme(HTMLAttributeValueOption):
    def __init__(self, name):
        super(BackgroundScheme, self).__init__(name, name is not None, 
                                               prefix='bg', constrain_value_to=['primary', 'dark', 'light'])


class NavbarLayout(Layout):
    """Used to populate a Navbar.

    :param fixed_to: May be one of 'top','bottom' or 'stickytop'.
                    The Navbar will stick to the top or bottom of the viewport.
    :param center_contents: If True, all the contents of the Navbar is centered within the Navbar itself.
    :param colour_theme: Use 'light' for use with light background colors, or 'dark' with dark background colors.
    :param bg_scheme: Whether the Navbar should use 'primary' colors, a 'dark' (light on dark) scheme
                        or a 'light' background.
    """
    def __init__(self, fixed_to=None, center_contents=False, colour_theme=None, bg_scheme=None):
        super(NavbarLayout, self).__init__()

        self.fixed = NavbarFixed(fixed_to)
        self.center_contents = center_contents
        self.colour_theme = ColourTheme(colour_theme)
        self.bg_scheme = BackgroundScheme(bg_scheme)
        self.brand = None
        self.toggle = None
        self.contents_container = None

    @property
    def nav(self):
        return self.widget.html_representation

    def customise_widget(self):
        super(NavbarLayout, self).customise_widget()
        if self.center_contents:
            centering_div = self.nav.add_child(Div(self.view).use_layout(Container()))
            self.contents_container = centering_div
        else:
            self.contents_container = self.nav
        self.main_container = self.contents_container

        if self.fixed.is_set:
            self.widget.append_class(self.fixed.as_html_snippet())

        for option in [self.colour_theme, self.bg_scheme]:
            if option.is_set:
                self.nav.append_class(option.as_html_snippet())

    def set_brand_text(self, brand_text):
        """Sets the brand to be a link to the home page that contains the given text.

        :param brand_text: Text to use for branding.
        """
        brand_a = A(self.view, Url('#'), description=brand_text)
        self.set_brand(brand_a)

    def insert_brand_widget(self, brand_html_element):
        index = 1 if self.toggle else 0
        self.main_container.insert_child(index, brand_html_element)

    @arg_checks(brand_html_element=IsInstance(HTMLWidget))
    def set_brand(self, brand_htmlwidget):
        """Sets `brand_widget` to be used as branding.

        :param brand_htmlwidget: An :class:`~reahl.web.ui.HTMLWidget` to be used as branding.
        """
        if self.brand:
            raise ProgrammerError('Brand has already been set to: %s' % self.brand)

        self.insert_brand_widget(brand_htmlwidget)
        brand_htmlwidget.append_class('navbar-brand')
        self.brand = brand_htmlwidget
        return self.brand

    @arg_checks(widget=IsInstance((reahl.web.bootstrap.navs.Nav, Form, TextNode)))
    def add(self, widget):
        """Adds the given Form or Nav `widget` to the Navbar.

        :param widget: A :class:`~reahl.web.bootstrap.navs.Nav`, :class:`~reahl.web.bootstrap.ui.Form` or
                       :class:`~reahl.web.bootstrap.ui.TextNode` to add.
        """
        if isinstance(widget, reahl.web.bootstrap.navs.Nav):
            widget.append_class('navbar-nav')
        if isinstance(widget, Form):
            widget.append_class('form-inline')
        if isinstance(widget, TextNode):
            span = Span(self.view)
            span.add_child(widget)
            span.append_class('navbar-text')
            widget = span
        return self.contents_container.add_child(widget)

    def add_toggle(self, target_html_element, text=None, left_aligned=False):
        """Adds a link that toggles the display of the given `target_html_element`.

        :param target_html_element: A :class:`~reahl.web.ui.HTMLElement`
        :param text: Text to be used on the toggle link. If None, the boostrap navbar-toggler-icon is used
        :keyword left_aligned: If True, ensure that the toggle is to the far left.
        """
        if not target_html_element.css_id_is_set:
            raise ProgrammerError('%s has no css_id set. A toggle is required to have a css_id' % target_html_element)
        target_html_element.append_class('collapse')
        toggle = CollapseToggle(self.view, target_html_element, text=text)
        index = 1 if (self.brand and not left_aligned) else 0
        self.main_container.insert_child(index, toggle)

        self.toggle = toggle
        return toggle


class ResponsiveLayout(NavbarLayout):
    """Makes the Navbar responsive and collapseable depending on the device.


    :param collapse_below_device_class: Which device class should trigger responsive collapsing.
                                        Valid values: xs, sm, md, lg, xl
    :param fixed_to: May be one of 'fixed-top','fixed-bottom' or 'sticky-top'.
                    The Navbar will stick to the top or bottom of the viewport.
    :param center_contents: If True, all the contents of the Navbar is centered within the Navbar itself.
    :param align_toggle_left: If True align the toggle on the left (the default is to align it right).
    :param collapse_brand_with_content: When set to True, the brand should collapse with the content.
    :param colour_theme: Use 'light' for use with light background colors, or 'dark' with dark background colors.
    :param bg_scheme: Whether the Navbar should use 'primary' colors, a 'dark' (light on dark) scheme
                        or a 'light' background.
    :param text: Text to be used on the toggle link. If None, the boostrap navbar-toggler-icon is used.

    """
    def __init__(self, collapse_below_device_class, center_contents=False, fixed_to=None, align_toggle_left=False,
                 collapse_brand_with_content=False,
                 colour_theme=None, bg_scheme=None, text=None):
        super(ResponsiveLayout, self).__init__(fixed_to=fixed_to, center_contents=center_contents,
                                               colour_theme=colour_theme, bg_scheme=bg_scheme)
        self.collapse_below_device_class = DeviceClass(collapse_below_device_class)
        self.collapse_brand_with_content = collapse_brand_with_content
        self.align_toggle_left = align_toggle_left
        if not self.collapse_below_device_class.one_smaller:
            raise ProgrammerError(('There is no device class smaller than %s' %
                                   self.collapse_below_device_class) +
                                  ' It does not make sense to collapse only '
                                  'if the viewport is smaller than the smallest device')
        self.text = text
        self.collapsing_content = None

    def insert_brand_widget(self, brand_html_element):
        if self.collapse_brand_with_content:
            self.collapsing_content.insert_child(0, brand_html_element)
        else:
            # getting in here also means self.contents_container is NOT the centering_div as it would be in the other leg
            if self.align_toggle_left:
                super(ResponsiveLayout, self).insert_brand_widget(brand_html_element)
            else:
                self.main_container.insert_child(0, brand_html_element)


    def customise_widget(self):
        super(ResponsiveLayout, self).customise_widget()
        if not self.widget.css_id_is_set:
            raise ProgrammerError('%s has no css_id set. A %s can only be used with a Widget that has a css_id' %
                                  (self.widget, self.__class__))

        collapsing_content = Div(self.view, css_id='%s_collapsable' % self.widget.css_id)
        collapsing_content.append_class('navbar-collapse')
        self.collapsing_content = collapsing_content
        self.contents_container.add_child(collapsing_content)
        self.contents_container = collapsing_content

        self.add_toggle(collapsing_content, text=self.text, left_aligned=self.align_toggle_left)

        toggle_size = self.collapse_below_device_class.one_smaller
        self.nav.append_class('navbar-expand-%s' % toggle_size.name)



class Navbar(HTMLWidget):
    """A bootstrap Navbar can be used as page header in a web
    application. A Navbar can contain some branding text, a
    :class:`~reahl.web.bootstrap.navs.Nav` and a small
    :class:`~reahl.web.bootstrap.forms.Form`.

    You populate a Navbar using a NavbarLayout for complete control
    over how the Navbar is laid out, which of its possible elements 
    are present and how they are composed themselves.

    :param view: (See :class:`reahl.web.fw.Widget`)
    :param css_id: (See :class:`reahl.web.fw.Widget`)
    """
    def __init__(self, view, css_id=None):
        super(Navbar, self).__init__(view)

        self.navbar = self.add_child(Nav(view))
        self.navbar.append_class('navbar')
        self.set_html_representation(self.navbar)
        if css_id:
            self.set_id(css_id)



# Copyright 2015-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. versionadded:: 3.2

Bootstrap-styled versions of Forms, Inputs and related Layouts.



"""
from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.component.exceptions import arg_checks, IsInstance
from reahl.component.i18n import Catalogue
from reahl.component.modelinterface import BooleanField, MultiChoiceField, Choice, Field

import reahl.web.ui
from reahl.web.ui import Label, HTMLAttributeValueOption
from reahl.web.bootstrap.ui import Div, P, WrappedInput, A, TextNode, Span, Legend, FieldSet
from reahl.web.bootstrap.grid import ColumnLayout


_ = Catalogue('reahl-web')


class Form(reahl.web.ui.Form):
    """A Form is a container for Inputs. Any Input has to belong to a Form. When a user clicks on
       a Button associated with a Form, the Event to which the Button is linked occurs at the
       server. The value of every Input that is associated with the Form is sent along with the
       Event to the server.

       :param view: (See :class:`reahl.web.fw.Widget`)
       :param unique_name: A name for this form, unique in the UserInterface where it is used.
    """
    javascript_widget_name = 'bootstrapform'


class NestedForm(reahl.web.ui.NestedForm):
    """A NestedForm can create the appearance of one Form being visually contained in
       another. Forms are not allowed to be children of other Forms but this restriction does
       not apply to NestedForms.

       :param view: (See :class:`reahl.web.fw.Widget`)
       :param unique_name: (See :class:`Form`)
       :keyword css_id: (See :class:`HTMLElement`)

    """
    def create_out_of_bound_form(self, view, unique_name):
        return Form(view, unique_name, rendered_form=self)


class TextInput(reahl.web.ui.TextInput):
    """A single line Input for typing plain text.

       :param form: (See :class:`~reahl.web.ui.Input`)
       :param bound_field: (See :class:`~reahl.web.ui.Input`)
       :param fuzzy: If True, the typed input will be dealt with as "fuzzy input". Fuzzy input is
                     when a user is allowed to type almost free-form input for structured types of input,
                     such as a date. The assumption is that the `bound_field` used should be able to parse
                     such "fuzzy input". If fuzzy=True, the typed value will be changed on the fly to
                     the system's interpretation of what the user originally typed as soon as the TextInput
                     looses focus.
       :param placeholder: If given a string, placeholder is displayed in the TextInput if the TextInput
                     is empty in order to provide a hint to the user of what may be entered into the TextInput.
                     If given True instead of a string, the label of the TextInput is used.
    """
    def __init__(self, form, bound_field, fuzzy=False, placeholder=False):
        super(TextInput, self).__init__(form, bound_field, fuzzy=fuzzy, placeholder=placeholder)
        self.append_class('form-control')


class PasswordInput(reahl.web.ui.PasswordInput):
    """A PasswordInput is a single line text input, but it does not show what the user is typing.

       :param form: (See :class:`~reahl.web.ui.Input`)
       :param bound_field: (See :class:`~reahl.web.ui.Input`)
    """
    def __init__(self, form, bound_field):
        super(PasswordInput, self).__init__(form, bound_field)
        self.append_class('form-control')


class TextArea(reahl.web.ui.TextArea):
    """A muli-line Input for plain text.

       :param form: (See :class:`~reahl.web.ui.Input`)
       :param bound_field: (See :class:`~reahl.web.ui.Input`)
       :param rows: The number of rows that this Input should have.
       :param columns: The number of columns that this Input should have.
    """
    def __init__(self, form, bound_field, rows=None, columns=None):
        super(TextArea, self).__init__(form, bound_field, rows=rows, columns=columns)
        self.append_class('form-control')


class SelectInput(reahl.web.ui.SelectInput):
    """An Input that lets the user select an :class:`reahl.component.modelinterface.Choice` from a dropdown
       list of valid ones.

       :param form: (See :class:`~reahl.web.ui.Input`)
       :param bound_field: (See :class:`~reahl.web.ui.Input`)
    """
    def __init__(self, form, bound_field):
        super(SelectInput, self).__init__(form, bound_field)
        self.append_class('custom-select')


PrimitiveCheckboxInput = reahl.web.ui.CheckboxInput


class SingleChoice(reahl.web.ui.SingleChoice):
    def create_html_widget(self):
        return self.create_button_input()

    @property
    def html_control(self):
        return self.html_representation


class CheckboxInput(reahl.web.ui.CheckboxSelectInput):
    """An Input that presents either a single checkbox or a list of them, depending on what Field
       it is used with.

       If used with a MultiChoiceField, the checkboxes represent which Choices are chosen; a single
       checkbox represents a BooleanField.

       :param form: (See :class:`~reahl.web.ui.Input`)
       :param bound_field: (See :class:`~reahl.web.ui.Input`)
       :param contents_layout: An optional :class:`ChoicesLayout` used to lay out the checkboxes in this input.
    """
    def __init__(self, form, bound_field, contents_layout=None):
        self.contents_layout = contents_layout
        self.checkbox_input = None
        super(CheckboxInput, self).__init__(form, bound_field)

    def create_html_widget(self):
        if isinstance(self.bound_field, BooleanField):
            main_element = self.create_main_element()
            self.added_choices.append(self.add_choice_to(main_element, Choice(self.bound_field.true_value, Field(label=self.bound_field.label))))
        else:
            main_element = super(CheckboxInput, self).create_html_widget()
        return main_element

    @property
    def includes_label(self):
        return not self.bound_field.allows_multiple_selections

    @property
    def jquery_selector(self):
        return '%s.closest("div")' % self.html_control.jquery_selector

    def create_main_element(self):
        return super(CheckboxInput, self).create_main_element().use_layout(self.contents_layout or ChoicesLayout(inline=False))

    def add_choice_to(self, widget, choice):
        single_choice = SingleChoice(self, choice)
        widget.layout.add_choice(single_choice)
        return single_choice


class RadioButtonSelectInput(reahl.web.ui.RadioButtonSelectInput):
    """An Input that lets the user select a :class:`reahl.component.modelinterface.Choice` from a list of valid ones
       shown as radio buttons of which only one can be selected at a time.

       :param form: (See :class:`~reahl.web.ui.Input`)
       :param bound_field: (See :class:`~reahl.web.ui.Input`)
       :param contents_layout: An optional :class:`ChoicesLayout` used to lay out the many choices in this input.
    """
    def __init__(self, form, bound_field, contents_layout=None):
        assert contents_layout is None or isinstance(contents_layout, ChoicesLayout), 'contents_layout should be an instance of ChoicesLayout but isn\'t' #TODO: this should be in @argchecks(...)
        self.contents_layout = contents_layout or ChoicesLayout(inline=False)
        super(RadioButtonSelectInput, self).__init__(form, bound_field)

    def create_main_element(self):
        main_element = super(RadioButtonSelectInput, self).create_main_element().use_layout(self.contents_layout)
        return main_element

    def add_choice_to(self, widget, choice):
        return widget.layout.add_choice(SingleChoice(self, choice))


class ButtonInput(reahl.web.ui.ButtonInput):
    """A button.

       :param form: (See :class:`~reahl.web.ui.Input`)
       :param event: The :class:`~reahl.web.component.modelinterface.Event` that will fire when the user clicks on this ButtonInput.
       :keyword css_id: (See :class:`HTMLElement`)
    """
    def __init__(self, form, event):
        super(ButtonInput, self).__init__(form, event)
        self.append_class('btn')

Button = ButtonInput


class StaticData(reahl.web.ui.Input):
    """A fake input which just displays the value of the :class:`~reahl.component.modelinterface.Field`
       to which the StaticData is attached, but does not include a way to change the value.

       This is useful in cases where you want to display some data that is exposed via a
       :class:`~reahl.component.modelinterface.Field` in a Form amongst normal Inputs.

       :param form: (See :class:`~reahl.web.ui.Input`)
       :param bound_field: (See :class:`~reahl.web.ui.Input`)
    """
    def __init__(self, form, bound_field):
        super(StaticData, self).__init__(form, bound_field)
        p = self.add_child(P(self.view, text=self.value))
        p.append_class('form-control-static')

    def can_write(self):
        return False


class CueInput(reahl.web.ui.WrappedInput):
    """A Widget that wraps around a given Input to augment it with a "cue" - a hint that
       appears only when the Input has focus. The intention of the cue is to give the
       user a hint as to what to input into the Input.

       :param html_input: The :class:`~reahl.web.ui.Input` to be augmented.
       :param cue_widget: An :class:`~reahl.web.fw.Widget` that serves as the cue.
    """
    def __init__(self, html_input, cue_widget):
        super(CueInput, self).__init__(html_input)
        div = self.add_child(Div(self.view))
        self.set_html_representation(div)

        div.append_class('reahl-bootstrapcueinput')
        cue_widget.append_class('reahl-bootstrapcue')

        div.add_child(html_input)
        self.cue_widget = div.add_child(cue_widget)

    def get_js(self, context=None):
        js = ['$(".reahl-bootstrapcueinput").bootstrapcueinput();']
        return super(CueInput, self).get_js(context=context) + js

    @property
    def includes_label(self):
        return self.input_widget.includes_label


class ButtonStyle(HTMLAttributeValueOption):
    valid_options = ['primary', 'secondary', 'success', 'info', 'warning', 'danger', 'link', 'light', 'dark']
    def __init__(self, name, outline=False):
        super(ButtonStyle, self).__init__(name, name is not None, prefix='btn' if not outline else 'btn-outline',
                                          constrain_value_to=self.valid_options)


class ButtonSize(HTMLAttributeValueOption):
    valid_options = ['lg', 'sm', 'xs']
    def __init__(self, size_string):
        super(ButtonSize, self).__init__(size_string, size_string is not None, prefix='btn',
                                         constrain_value_to=self.valid_options)


class ButtonLayout(reahl.web.fw.Layout):
    """A ButtonLayout can be used to make something (like an :class:`A`) look like
       a :class:`Button`. It has a few options controlling specifics of that look,
       and can be used to change the default look of a :class:`Button` as well.

       :keyword style: The general style of the button
                   (one of: 'default', 'primary', 'success', 'info', 'warning', 'danger', 'link', 'light', 'dark')
       :keyword size: The size of the button (one of: 'xs', 'sm', 'lg')
       :keyword active: If True, the button is visually altered to indicate it is active
                        (buttons can be said to be active in the same sense that a menu item can
                        be the currently active menu item).
       :keyword wide: If True, the button stretches to the entire width of its parent.

    """
    def __init__(self, style=None, outline=False, size=None, active=False, wide=False):
        super(ButtonLayout, self).__init__()
        self.style = ButtonStyle(style, outline=outline)
        self.size = ButtonSize(size)
        self.active = HTMLAttributeValueOption('active', active)
        self.wide = HTMLAttributeValueOption('btn-block', wide)

    def customise_widget(self):
        self.widget.append_class('btn')

        if isinstance(self.widget, A) and self.widget.disabled:
            self.widget.append_class('disabled')
        for option in [self.style, self.size, self.active, self.wide]:
            if option.is_set:
                self.widget.append_class(option.as_html_snippet())


class ChoicesLayout(reahl.web.fw.Layout):
    def __init__(self, inline=False):
        super(ChoicesLayout, self).__init__()
        self.inline = inline

    @arg_checks(html_input=IsInstance((PrimitiveCheckboxInput, SingleChoice)))
    def add_choice(self, html_input):
        input_type_custom_control = HTMLAttributeValueOption(html_input.choice_type, True, prefix='custom',
                                                             constrain_value_to=['radio', 'checkbox'])

        label_widget = Label(self.view, for_input=html_input)
        label_widget.append_class('custom-control-label')

        html_input.append_class('custom-control-input')

        outer_div = Div(self.view)
        outer_div.append_class('custom-control')
        outer_div.append_class(input_type_custom_control.as_html_snippet())
        if self.inline:
            outer_div.append_class('custom-control-inline')
        if html_input.disabled:
            outer_div.append_class('disabled')

        outer_div.add_child(html_input)
        outer_div.add_child(label_widget)

        self.widget.add_child(outer_div)

        return outer_div


class FormLayout(reahl.web.fw.Layout):
    """A FormLayout is used to create Forms that have a consistent look by arranging
       all its Inputs, their Labels and possible validation error messages in a
       certain way.

       This basic FormLayout positions Labels above added Inputs and allow for an
       optional helpful text message with each input. Validation error messages are
       displayed underneath invalid Inputs.

       Different kinds of FormLayouts allow different kinds of arrangements.

       A FormLayout need not be applied directly to a Form itself. It
       can also be applied to, say, a Div (or FieldSet) which is a child of a
       Form. This makes the arrangement quite flexible, since you
       could have different parts of a Form that are laid out using
       different FormLayouts or even by different types of FormLayout.
    """
    def create_form_group(self, html_input):
        if isinstance(html_input, RadioButtonSelectInput):
            form_group = self.widget.add_child(FieldSet(self.view))
        else:
            form_group = self.widget.add_child(Div(self.view))
        form_group.append_class('form-group')
        html_input.add_attribute_source(reahl.web.ui.ValidationStateAttributes(html_input,
                                                             error_class='is-invalid',
                                                             success_class='is-valid'))
        return form_group

    def add_input_to(self, parent_element, html_input):
        return parent_element.add_child(html_input)

    def add_validation_error_to(self, form_group, html_input):
        error_text = form_group.add_child(Span(self.view, text=html_input.validation_error.message))
        error_text.append_class('invalid-feedback')
        error_text.set_attribute('for', html_input.name) #need for our custom fileupload js
        error_text.set_attribute('generated', 'true')
        return error_text

    def create_help_text_widget(self, help_text):
        help_text_widget = P(self.view, text=help_text)
        help_text_widget.append_class('form-text')
        return help_text_widget

    def add_help_text_to(self, parent_element, html_input, help_text):
        help_text_widget = parent_element.add_child(self.create_help_text_widget(help_text))
        help_text_widget.append_class('text-muted')
        if not help_text_widget.css_id_is_set:
            help_text_widget.generate_random_css_id()
        html_input.set_attribute('aria-describedby', help_text_widget.css_id)
        return help_text_widget

    def add_label_to(self, form_group, html_input, hidden):
        if isinstance(html_input, RadioButtonSelectInput):
            label = form_group.add_child(Legend(self.view, text=html_input.label))
            label.append_class('col-form-label')
        else:
            label = form_group.add_child(Label(self.view, for_input=html_input))
        if hidden:
            label.append_class('sr-only')
        return label

    def add_input(self, html_input, hide_label=False, help_text=None):
        """Adds an input to the Form.

           :param html_input: The Input to add.
           :keyword hide_label: If True, makes the label invisible yet available to screenreaders.
           :keyword help_text: Helpful text to display with each Input field.
        """
        form_group = self.create_form_group(html_input)

        if not html_input.includes_label:
            self.add_label_to(form_group, html_input, hide_label)

        self.add_input_to(form_group, html_input)

        if html_input.get_input_status() == 'invalidly_entered':
            self.add_validation_error_to(form_group, html_input)
            html_input.append_class('is-invalid')
        if html_input.get_input_status() == 'validly_entered':
            html_input.append_class('is-valid')

        if help_text:
            self.add_help_text_to(form_group, html_input, help_text)

        return html_input


class GridFormLayout(FormLayout):
    """A GridFormLayout arranges its Labels and Inputs in a grid with two columns. Labels
       go into the left column and Inputs into the right column. The programmer specifies
       how wide each column should be.

       :param label_column_size: A :class:`~reahl.web.bootstrap.grid.ResponsiveSize` for the width of the Label column.
       :param input_column_size: A :class:`~reahl.web.bootstrap.grid.ResponsiveSize` for the width of the Input column.
    """
    def __init__(self, label_column_size, input_column_size):
        super(GridFormLayout, self).__init__()
        self.label_column_size = label_column_size
        self.input_column_size = input_column_size

    def create_form_group(self, html_input):
        form_group = super(GridFormLayout, self).create_form_group(html_input)
        form_group.use_layout(ColumnLayout())
        form_group.layout.add_column('label', size=self.label_column_size)
        form_group.layout.add_column('input', size=self.input_column_size)
        return form_group

    def add_label_to(self, form_group, html_input, hidden):
        column = form_group.layout.columns['label']
        label = super(GridFormLayout, self).add_label_to(column, html_input, hidden)
        label.append_class('col-form-label')
        return label

    def add_input_to(self, parent_element, html_input):
        input_column = parent_element.layout.columns['input']
        return super(GridFormLayout, self).add_input_to(input_column, html_input)

    def add_help_text_to(self, parent_element, html_input, help_text):
        input_column = parent_element.layout.columns['input']
        return super(GridFormLayout, self).add_help_text_to(input_column, html_input, help_text)


class InlineFormLayout(FormLayout):
    """A FormLayout which positions all its Inputs and Labels on one line. The browser
       flows this like any paragraph of text. Each Label precedes its associated Input."""
    def customise_widget(self):
        super(InlineFormLayout, self).customise_widget()
        self.widget.append_class('form-inline')

    def create_help_text_widget(self, help_text):
        return Span(self.view, text=help_text)

    def add_label_to(self, form_group, html_input, hidden):
        label = super(InlineFormLayout, self).add_label_to(form_group, html_input, hidden)
        label.append_class('mr-2')
        return label

    def add_help_text_to(self, parent_element, html_input, help_text):
        help_text = super(InlineFormLayout, self).add_help_text_to(parent_element, html_input, help_text)
        help_text.append_class('ml-2')
        return help_text


class InputGroup(reahl.web.ui.WrappedInput):
    """A composition of an Input with something preceding and/or following it.

    :param prepend: A :class:`~reahl.web.fw.Widget` or text to prepend to the :class:`~reahl.web.ui.Input`.
    :param input_widget: The :class:`~reahl.web.ui.Input` to use.
    :param append: A :class:`~reahl.web.fw.Widget` or text to append to the :class:`~reahl.web.ui.Input`.
    """
    def __init__(self, prepend, input_widget, append):
        super(InputGroup, self).__init__(input_widget)
        self.div = self.add_child(Div(self.view))
        self.div.append_class('input-group')
        if prepend:
            self.add_as_addon(prepend, 'prepend')
        self.input_widget = self.div.add_child(input_widget)
        if append:
            self.add_as_addon(append, 'append')
        self.set_html_representation(self.div)

    def add_as_addon(self, addon, position):
        if isinstance(addon, six.string_types):
            addon = Span(self.view, text=addon)
        span = Span(self.view)
        span.add_child(addon)
        span.append_class('input-group-%s' % position)
        addon.append_class('input-group-text')
        return self.div.add_child(span)
